import importlib
import logging
import os

from functools import wraps
from flask import abort, request, redirect, Response
from redis import Redis

# NOTE: The HTTP service will *only* work on top of a Redis bus. The default
# internal bus service won't work as the web server will run in a different process.
from platypush.bus.redis import RedisBus

from platypush.config import Config
from platypush.message import Message
from platypush.message.request import Request
from platypush.user import UserManager
from platypush.utils import get_redis_queue_name_by_message, get_ip_or_hostname

_bus = None
_logger = None


def bus():
    global _bus
    if _bus is None:
        _bus = RedisBus()
    return _bus


def logger():
    global _logger
    if not _logger:
        log_args = {
            'level': logging.INFO,
            'format': '%(asctime)-15s|%(levelname)5s|%(name)s|%(message)s',
        }

        level = (Config.get('backend.http') or {}).get('logging') or \
            (Config.get('logging') or {}).get('level')
        filename = (Config.get('backend.http') or {}).get('filename')

        if level:
            log_args['level'] = getattr(logging, level.upper()) \
                if isinstance(level, str) else level
        if filename:
            log_args['filename'] = filename

        logging.basicConfig(**log_args)
        _logger = logging.getLogger('platyweb')

    return _logger


def get_message_response(msg):
    redis = Redis(**bus().redis_args)
    response = redis.blpop(get_redis_queue_name_by_message(msg), timeout=60)
    if response and len(response) > 1:
        response = Message.build(response[1])
    else:
        response = None

    return response


# noinspection PyProtectedMember
def get_http_port():
    from platypush.backend.http import HttpBackend
    http_conf = Config.get('backend.http')
    return http_conf.get('port', HttpBackend._DEFAULT_HTTP_PORT)


# noinspection PyProtectedMember
def get_websocket_port():
    from platypush.backend.http import HttpBackend
    http_conf = Config.get('backend.http')
    return http_conf.get('websocket_port', HttpBackend._DEFAULT_WEBSOCKET_PORT)


def send_message(msg, wait_for_response=True):
    msg = Message.build(msg)

    if isinstance(msg, Request):
        msg.origin = 'http'

    if Config.get('token'):
        msg.token = Config.get('token')

    bus().post(msg)

    if isinstance(msg, Request) and wait_for_response:
        response = get_message_response(msg)
        logger().debug('Processing response on the HTTP backend: {}'.
                       format(response))

        return response


def send_request(action, wait_for_response=True, **kwargs):
    msg = {
        'type': 'request',
        'action': action
    }

    if kwargs:
        msg['args'] = kwargs

    return send_message(msg, wait_for_response=wait_for_response)


def _authenticate_token():
    token = Config.get('token')

    if 'X-Token' in request.headers:
        user_token = request.headers['X-Token']
    elif 'token' in request.args:
        user_token = request.args.get('token')
    else:
        return False

    return token and user_token == token


def _authenticate_http():
    user_manager = UserManager()

    if not request.authorization:
        return False

    username = request.authorization.username
    password = request.authorization.password
    return user_manager.authenticate_user(username, password)


def _authenticate_session():
    user_manager = UserManager()
    user_session_token = None
    user = None

    if 'X-Session-Token' in request.headers:
        user_session_token = request.headers['X-Session-Token']
    elif 'session_token' in request.args:
        user_session_token = request.args.get('session_token')
    elif 'session_token' in request.cookies:
        user_session_token = request.cookies.get('session_token')

    if user_session_token:
        user, session = user_manager.authenticate_user_session(user_session_token)

    return user is not None


def _authenticate_csrf_token():
    user_manager = UserManager()
    user_session_token = None

    if 'X-Session-Token' in request.headers:
        user_session_token = request.headers['X-Session-Token']
    elif 'session_token' in request.args:
        user_session_token = request.args.get('session_token')
    elif 'session_token' in request.cookies:
        user_session_token = request.cookies.get('session_token')

    if user_session_token:
        user, session = user_manager.authenticate_user_session(user_session_token)
    else:
        return False

    if user is None:
        return False

    return session.csrf_token is None or request.form.get('csrf_token') == session.csrf_token


def authenticate(redirect_page='', skip_auth_methods=None, check_csrf_token=False):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            user_manager = UserManager()
            n_users = user_manager.get_user_count()
            token = Config.get('token')
            skip_methods = skip_auth_methods or []

            # User/pass HTTP authentication
            http_auth_ok = True
            if n_users > 0 and 'http' not in skip_methods:
                http_auth_ok = _authenticate_http()
                if http_auth_ok:
                    return f(*args, **kwargs)

            # Token-based authentication
            token_auth_ok = True
            if token and 'token' not in skip_methods:
                token_auth_ok = _authenticate_token()
                if token_auth_ok:
                    return f(*args, **kwargs)

            # Session token based authentication
            session_auth_ok = True
            if n_users > 0 and 'session' not in skip_methods:
                session_auth_ok = _authenticate_session()
                if session_auth_ok:
                    return f(*args, **kwargs)

                return redirect('/login?redirect=' + redirect_page, 307)

            # CSRF token check
            if check_csrf_token:
                csrf_check_ok = _authenticate_csrf_token()
                if not csrf_check_ok:
                    return abort(403, 'Invalid or missing csrf_token')

            if n_users == 0 and 'session' not in skip_methods:
                return redirect('/register?redirect=' + redirect_page, 307)

            if ('http' not in skip_methods and http_auth_ok) or \
                    ('token' not in skip_methods and token_auth_ok) or \
                    ('session' not in skip_methods and session_auth_ok):
                return f(*args, **kwargs)

            return Response('Authentication required', 401,
                            {'WWW-Authenticate': 'Basic realm="Login required"'})

        return wrapper

    return decorator


def get_routes():
    routes_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'routes')
    routes = []
    base_module = '.'.join(__name__.split('.')[:-1])

    for path, dirs, files in os.walk(routes_dir):
        for f in files:
            if f.endswith('.py'):
                mod_name = '.'.join(
                    (base_module + '.' + os.path.join(path, f).replace(
                        os.path.dirname(__file__), '')[1:].replace(os.sep, '.')).split('.')
                    [:(-2 if f == '__init__.py' else -1)])

                try:
                    mod = importlib.import_module(mod_name)
                    if hasattr(mod, '__routes__'):
                        routes.extend(mod.__routes__)
                except Exception as e:
                    logger().warning('Could not import routes from {}/{}: {}: {}'.
                                     format(path, f, type(e), str(e)))

    return routes


def get_local_base_url():
    http_conf = Config.get('backend.http') or {}
    return '{proto}://localhost:{port}'.format(
        proto=('https' if http_conf.get('ssl_cert') else 'http'),
        port=get_http_port())


def get_remote_base_url():
    http_conf = Config.get('backend.http') or {}
    return '{proto}://{host}:{port}'.format(
        proto=('https' if http_conf.get('ssl_cert') else 'http'),
        host=get_ip_or_hostname(), port=get_http_port())


# vim:sw=4:ts=4:et:



import inspect
import logging
import threading


class EventGenerator(object):
    """
    Abstract class for event generators. It allows to fire events to the bus and
    trigger/register/unregister custom callback handlers associated to event
    types. Both plugins and backends extend this class.
    """

    logger = logging.getLogger(__name__)

    def __init__(self, *args, **kwargs):
        self._event_handlers = {}   # Event type => callback map

    def fire_event(self, event):
        """
        Fires an event (instance of :class:`platypush.message.event.Event` or a
            subclass) to the internal bus and triggers any handler callback
            associated to the event type or any of its super-classes.

        :param event: Event to fire
        :type event: :class:`platypush.message.event.Event` or a subclass
        """

        def hndl_thread():
            hndl(event)

        from platypush.backend import Backend
        from platypush.context import get_bus

        bus = self.bus if isinstance(self, Backend) else get_bus()
        if not bus:
            self.logger.warning('No bus available to post the event: {}'.format(event))
        else:
            bus.post(event)

        handlers = set()

        for cls in inspect.getmro(event.__class__):
            if cls in self._event_handlers:
                handlers.update(self._event_handlers[cls])

        for hndl in handlers:
            threading.Thread(target=hndl_thread).start()


    def register_handler(self, event_type, callback):
        """
        Registers a callback handler for a camera event type.

        :param event_type: Event type to listen to. Must be a subclass of
            :class:`platypush.message.event.Event`
        :type event_type: :class:`platypush.message.event.Event` or a subclass

        :param callback: Callback function. It will take an event of type
            :class:`platypush.message.event.Event` as a parameter
        :type callback: function
        """

        if event_type not in self._event_handlers:
            self._event_handlers[event_type] = set()
        self._event_handlers[event_type].add(callback)


    def unregister_handler(self, event_type, callback):
        """
        Unregisters a callback handler from a camera event type.

        :param event_type: Event type the callback is registered to
        :type event_type: :class:`platypush.message.event.Event` or a subclass

        :param callback: Callback function to unregister
        :type callback: function
        """

        if event_type not in self._event_handlers:
            return
        if callback not in self._event_handlers[event_type]:
            return

        self._event_handlers[event_type].remove(callback)

        if not self._event_handlers[event_type]:
            del self._event_handlers[event_type]


# vim:sw=4:ts=4:et:



import sys

class Logger:
    def __init__(self, level):
        self.level = level

    def write(self, message):
        if message is None:
            return

        if isinstance(message, bytes):
            message = message.decode()

        message = message.rstrip()
        if message and len(message) > 0:
            self.level(message)

    def flush(self):
        pass


# vim:sw=4:ts=4:et:



import datetime
from typing import Optional

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.travisci import TravisciBuildPassedEvent, TravisciBuildFailedEvent


class TravisciBackend(Backend):
    """
    This backend polls for new builds on a `Travis-Ci <https://travis-ci.org>`_ account and triggers an event
    whenever a new build is completed.

    Requires:

        * The :class:`platypush.plugins.foursquare.FoursquarePlugin` plugin configured and enabled.

    Triggers:

        - :class:`platypush.message.event.travisci.TravisciBuildPassedEvent` when the build of a project owned by
            the user passes.
        - :class:`platypush.message.event.travisci.TravisciBuildFailedEvent` when the build of a project owned by
            the user fails.

    """

    _last_build_finished_at_varname = '_travisci_last_build_finished_at'

    def __init__(self, poll_seconds: Optional[float] = 60.0, *args, **kwargs):
        """
        :param poll_seconds: How often the backend should check for new builds (default: one minute).
        """
        super().__init__(*args, poll_seconds=poll_seconds, **kwargs)
        self._last_build_finished_at = None

    @staticmethod
    def _convert_iso_date(d):
        if not d:
            return

        try:
            return float(d)
        except ValueError:
            pass

        if isinstance(d, str):
            if d.endswith('Z'):
                d = d[:-1] + '+00:00'
            d = datetime.datetime.fromisoformat(d)

        assert isinstance(d, datetime.datetime), 'Unexpected datetime type: ' + type(d)
        return d.timestamp()

    def __enter__(self):
        self._last_build_finished_at = self._convert_iso_date(
            get_plugin('variable').get(self._last_build_finished_at_varname).
            output.get(self._last_build_finished_at_varname) or 0)

        self.logger.info('Started Travis-CI backend')

    def loop(self):
        builds = get_plugin('travisci').builds(limit=1).output
        if not builds:
            return

        last_build = builds[0]
        last_build_finished_at = self._convert_iso_date(last_build.get('finished_at', 0))
        if not last_build_finished_at:
            return

        if self._last_build_finished_at and last_build_finished_at <= self._last_build_finished_at:
            return

        if last_build.get('state') == 'passed':
            evt_type = TravisciBuildPassedEvent
        elif last_build.get('state') == 'failed':
            evt_type = TravisciBuildFailedEvent
        else:
            return

        evt = evt_type(repository_id=last_build.get('repository', {}).get('id'),
                       repository_name=last_build.get('repository', {}).get('name'),
                       repository_slug=last_build.get('repository').get('slug'),
                       build_id=int(last_build.get('id')),
                       build_number=int(last_build.get('number')),
                       duration=last_build.get('duration'),
                       previous_state=last_build.get('previous_state'),
                       private=last_build.get('private'),
                       tag=last_build.get('tag'),
                       branch=last_build.get('branch', {}).get('name'),
                       commit_id=last_build.get('commit', {}).get('id'),
                       commit_sha=last_build.get('commit', {}).get('sha'),
                       commit_message=last_build.get('commit', {}).get('message'),
                       committed_at=self._convert_iso_date(
                           last_build.get('commit', {}).get('committed_at')),
                       created_by=last_build.get('created_by', {}).get('login'),
                       started_at=self._convert_iso_date(last_build.get('started_at')),
                       finished_at=self._convert_iso_date(last_build.get('finished_at')))

        self.bus.post(evt)
        self._last_build_finished_at = last_build_finished_at
        get_plugin('variable').set(**{self._last_build_finished_at_varname: self._last_build_finished_at})


# vim:sw=4:ts=4:et:



import time

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.plugins.stt import SttPlugin


class SttBackend(Backend):
    """
    Base class for speech-to-text backends.
    """

    def __init__(self, plugin_name: str, retry_sleep: float = 5.0, *args, **kwargs):
        """
        :param plugin_name: Plugin name of the class that will be used for speech detection. Must be an instance of
            :class:`platypush.plugins.stt.SttPlugin`.
        :param retry_sleep: Number of seconds the backend will wait on failure before re-initializing the plugin
            (default: 5 seconds).
        """
        super().__init__(*args, **kwargs)
        self.plugin_name = plugin_name
        self.retry_sleep = retry_sleep

    def run(self):
        super().run()
        self.logger.info('Starting {} speech-to-text backend'.format(self.__class__.__name__))

        while not self.should_stop():
            try:
                plugin: SttPlugin = get_plugin(self.plugin_name)
                with plugin:
                    # noinspection PyProtectedMember
                    plugin._detection_thread.join()
            except Exception as e:
                self.logger.exception(e)
                self.logger.warning('Encountered an unexpected error, retrying in {} seconds'.format(self.retry_sleep))
                time.sleep(self.retry_sleep)


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

from platypush.backend.assistant.google import AssistantGoogleBackend
from platypush.context import get_backend
from platypush.plugins import action
from platypush.plugins.assistant import AssistantPlugin


class AssistantGooglePlugin(AssistantPlugin):
    """
    Google assistant plugin.
    It acts like a wrapper around the :mod:`platypush.backend.assistant.google`
    backend to programmatically control the conversation status.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _get_assistant(self) -> AssistantGoogleBackend:
        return get_backend('assistant.google')

    @action
    def start_conversation(self, **kwargs):
        """
        Programmatically start a conversation with the assistant
        """
        assistant = self._get_assistant()
        assistant.start_conversation()

    @action
    def stop_conversation(self):
        """
        Programmatically stop a running conversation with the assistant
        """
        assistant = self._get_assistant()
        assistant.stop_conversation()

    @action
    def set_mic_mute(self, muted: bool = True):
        """
        Programmatically mute/unmute the microphone.

        :param muted: Set to True or False.
        """
        assistant = self._get_assistant()
        assistant.set_mic_mute(muted)

    @action
    def toggle_mic_mute(self):
        """
        Toggle the mic mute state.
        """
        assistant = self._get_assistant()
        is_muted = assistant.is_muted()
        self.set_mic_mute(muted=not is_muted)

    @action
    def is_muted(self) -> bool:
        """
        :return: True if the microphone is muted, False otherwise.
        """
        assistant = self._get_assistant()
        return assistant.is_muted()

    @action
    def send_text_query(self, query: str):
        """
        Send a text query to the assistant.

        :param query: Query to be sent.
        """
        assistant = self._get_assistant()
        assistant.send_text_query(query)


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class LeapFrameEvent(Event):
    """
    Event triggered when a Leap Motion devices receives a new frame
    """

    def __init__(self, hands, *args, **kwargs):
        """
        :param hands: Reference to the detected hands properties (palm and fingers X,Y,Z position, direction etc.)
        :type hands: dict
        """

        super().__init__(hands=hands, *args, **kwargs)


class LeapFrameStartEvent(Event):
    """
    Event triggered when a new sequence of frames is detected by the Leap Motion sensor
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class LeapFrameStopEvent(Event):
    """
    Event triggered when a Leap Sensor stops detecting frames
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class LeapConnectEvent(Event):
    """
    Event triggered when a Leap Motion sensor is connected
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class LeapDisconnectEvent(Event):
    """
    Event triggered when a Leap Motion sensor is disconnected
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# vim:sw=4:ts=4:et:




import base64

from flask import abort, request, Blueprint, Response

from platypush.backend.http.app import template_folder
from platypush.context import get_plugin
from platypush.plugins.qrcode import QrcodePlugin

qrcode = Blueprint('qrcode', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    qrcode,
]


@qrcode.route('/qrcode', methods=['GET'])
def generate_code():
    """
    This route can be used to generate a QR code given a ``content`` parameter.
    """

    content = request.args.get('content')
    if not content:
        abort(400, 'Expected content parmeter')

    plugin: QrcodePlugin = get_plugin('qrcode')
    response = plugin.generate(content, format='png').output
    data = base64.decodebytes(response['data'].encode())
    return Response(data, mimetype='image/png')


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class DistanceSensorEvent(Event):
    """
    Event triggered when a new value is processed by a distance sensor.
    """
    def __init__(self, distance: float, unit: str = 'mm', *args, **kwargs):
        super().__init__(*args, distance=distance, unit=unit, disable_logging=True,
                         disable_web_clients_notification=True, **kwargs)


# vim:sw=4:ts=4:et:



import enum

from platypush.message.response.bluetooth import BluetoothScanResponse
from platypush.plugins import action
from platypush.plugins.bluetooth.ble import BluetoothBlePlugin
from platypush.plugins.switch import SwitchPlugin


class SwitchSwitchbotPlugin(SwitchPlugin, BluetoothBlePlugin):
    """
    Plugin to interact with a Switchbot (https://www.switch-bot.com/) device and
    programmatically control buttons.

    See :class:`platypush.plugins.bluetooth.ble.BluetoothBlePlugin` for how to enable BLE permissions for
    the platypush user (a simple solution may be to run it as root, but that's usually NOT a good idea).

    Requires:

        * **pybluez** (``pip install pybluez``)
        * **gattlib** (``pip install gattlib``)
        * **libboost** (on Debian ```apt-get install libboost-python-dev libboost-thread-dev``)
    """

    uuid = 'cba20002-224d-11e6-9fb8-0002a5d5c51b'
    handle = 0x16

    class Command(enum.Enum):
        """
        Base64 encoded commands
        """
        # \x57\x01\x00
        PRESS = 'VwEA'
        # # \x57\x01\x01
        ON = 'VwEB'
        # # \x57\x01\x02
        OFF = 'VwEC'

    def __init__(self, interface=None, connect_timeout=None,
                 scan_timeout=2, devices=None, **kwargs):
        """
        :param interface: Bluetooth interface to use (e.g. hci0) default: first available one
        :type interface: str

        :param connect_timeout: Timeout for the connection to the Switchbot device - default: None
        :type connect_timeout: float

        :param scan_timeout: Timeout for the scan operations
        :type scan_timeout: float

        :param devices: Devices to control, as a MAC address -> name map
        :type devices: dict
        """

        SwitchPlugin.__init__(self, **kwargs)
        BluetoothBlePlugin.__init__(self, interface=interface)

        self.connect_timeout = connect_timeout if connect_timeout else 5
        self.scan_timeout = scan_timeout if scan_timeout else 2
        self.configured_devices = devices or {}
        self.configured_devices_by_name = {
            name: addr
            for addr, name in self.configured_devices.items()
        }

    def _run(self, device: str, command: Command):
        if device in self.configured_devices_by_name:
            device = self.configured_devices_by_name[device]

        self.write(device, command.value, handle=self.handle, channel_type='random', binary=True)
        return self.status(device)

    @action
    def press(self, device):
        """
        Send a press button command to a device

        :param device: Device name or address
        :type device: str
        """
        return self._run(device, self.Command.PRESS)

    @action
    def toggle(self, device, **kwargs):
        return self.press(device)

    @action
    def on(self, device, **kwargs):
        """
        Send a press-on button command to a device

        :param device: Device name or address
        :type device: str
        """
        return self._run(device, self.Command.ON)

    @action
    def off(self, device, **kwargs):
        """
        Send a press-off button command to a device

        :param device: Device name or address
        :type device: str
        """
        return self._run(device, self.Command.OFF)

    @action
    def scan(self, interface: str = None, duration: int = 10) -> BluetoothScanResponse:
        """
        Scan for available Switchbot devices nearby.

        :param interface: Bluetooth interface to scan (default: default configured interface)
        :param duration: Scan duration in seconds
        """

        devices = super().scan(interface=interface, duration=duration).devices
        compatible_devices = {}

        for dev in devices:
            # noinspection PyBroadException
            try:
                characteristics = [
                    chrc for chrc in self.discover_characteristics(
                        dev['addr'], channel_type='random', wait=False,
                        timeout=self.scan_timeout).characteristics
                    if chrc.get('uuid') == self.uuid
                ]

                if characteristics:
                    compatible_devices[dev['addr']] = None
            except:
                pass

        return BluetoothScanResponse(devices=compatible_devices)

    @property
    def devices(self):
        return [
            {
                'address': addr,
                'id': addr,
                'name': name,
                'on': False,
            }
            for addr, name in self.configured_devices.items()
        ]


# vim:sw=4:ts=4:et:



import copy
import datetime
import json
import logging
import random
import re
import time

from threading import Thread

from platypush.config import Config
from platypush.context import get_plugin
from platypush.message import Message
from platypush.message.response import Response
from platypush.utils import get_hash, get_module_and_method_from_action, get_redis_queue_name_by_message

logger = logging.getLogger(__name__)


class Request(Message):
    """ Request message class """

    def __init__(self, target, action, origin=None, id=None, backend=None,
                 args=None, token=None, timestamp=None):
        """
        Params:
            target -- Target node [Str]
            action -- Action to be executed (e.g. music.mpd.play) [Str]
            origin -- Origin node [Str]
                id -- Message ID, or None to get it auto-generated
           backend -- Backend connected to the request, where the response will be delivered
              args -- Additional arguments for the action [Dict]
             token -- Authorization token, if required on the server [Str]
         timestamp -- Message creation timestamp [Float]
        """

        super().__init__(timestamp=timestamp)

        self.id = id if id else self._generate_id()
        self.target = target
        self.action = action
        self.origin = origin
        self.args = args if args else {}
        self.backend = backend
        self.token = token

    @classmethod
    def build(cls, msg):
        msg = super().parse(msg)
        args = {'target': msg.get('target', Config.get('device_id')), 'action': msg['action'],
                'args': msg.get('args', {}), 'id': msg['id'] if 'id' in msg else cls._generate_id(),
                'timestamp': msg['_timestamp'] if '_timestamp' in msg else time.time()}

        if 'origin' in msg:
            args['origin'] = msg['origin']
        if 'token' in msg:
            args['token'] = msg['token']
        return cls(**args)

    @staticmethod
    def _generate_id():
        _id = ''
        for i in range(0, 16):
            _id += '%.2x' % random.randint(0, 255)
        return _id

    def _execute_procedure(self, *args, **kwargs):
        from platypush.config import Config
        from platypush.procedure import Procedure

        logger.info('Executing procedure request: {}'.format(self.action))
        proc_name = self.action.split('.')[-1]
        proc_config = Config.get_procedures()[proc_name]
        proc = Procedure.build(name=proc_name, requests=proc_config['actions'],
                               _async=proc_config['_async'], args=self.args,
                               backend=self.backend, id=self.id)

        return proc.execute(*args, **kwargs)

    def _expand_context(self, event_args=None, **context):
        from platypush.config import Config

        if event_args is None:
            event_args = copy.deepcopy(self.args)

        constants = Config.get_constants()
        context['constants'] = {}
        for (name, value) in constants.items():
            context['constants'][name] = value

        keys = []
        if isinstance(event_args, dict):
            keys = event_args.keys()
        elif isinstance(event_args, list):
            keys = range(0, len(event_args))

        for key in keys:
            value = event_args[key]

            if isinstance(value, str):
                value = self.expand_value_from_context(value, **context)
            elif isinstance(value, dict) or isinstance(value, list):
                self._expand_context(event_args=value, **context)

            event_args[key] = value

        return event_args

    # noinspection PyBroadException
    @classmethod
    def expand_value_from_context(cls, _value, **context):
        for (k, v) in context.items():
            if isinstance(v, Message):
                v = json.loads(str(v))
            try:
                exec('{}={}'.format(k, v))
            except:
                if isinstance(v, str):
                    try:
                        exec('{}="{}"'.format(k, re.sub('(^|[^\\\])"', '\1\\"', v)))
                    except:
                        pass

        parsed_value = ''
        if not isinstance(_value, str):
            parsed_value = _value

        while _value and isinstance(_value, str):
            m = re.match('([^$]*)(\${\s*(.+?)\s*})(.*)', _value)
            if m and not m.group(1).endswith('\\'):
                prefix = m.group(1)
                expr = m.group(2)
                inner_expr = m.group(3)
                _value = m.group(4)

                try:
                    context_value = eval(inner_expr)

                    if callable(context_value):
                        context_value = context_value()
                    if isinstance(context_value, range) or isinstance(context_value, tuple):
                        context_value = [*context_value]
                    if isinstance(context_value, datetime.date):
                        context_value = context_value.isoformat()
                except Exception as e:
                    logger.exception(e)
                    context_value = expr

                parsed_value += prefix + (
                    json.dumps(context_value)
                    if isinstance(context_value, list) or isinstance(context_value, dict)
                    else str(context_value)
                )
            else:
                parsed_value += _value
                _value = ''

        try:
            return json.loads(parsed_value)
        except:
            return parsed_value

    def _send_response(self, response):
        response = Response.build(response)
        response.id = self.id
        response.target = self.origin
        response.origin = Config.get('device_id')

        if self.backend and self.origin:
            self.backend.send_response(response=response, request=self)
        else:
            redis = get_plugin('redis')
            if redis:
                queue_name = get_redis_queue_name_by_message(self)
                redis.send_message(queue_name, response)
                redis.expire(queue_name, 60)

    def execute(self, n_tries=1, _async=True, **context):
        """
        Execute this request and returns a Response object
        Params:
            n_tries -- Number of tries in case of failure before raising a RuntimeError
            _async   -- If True, the request will be run asynchronously and the
                       response posted on the bus when available (default),
                       otherwise the current thread will wait for the response
                       to be returned synchronously.
            context -- Key-valued context. Example:
                context = (group_name='Kitchen lights')
                request.args:
                    - group: ${group_name}  # will be expanded as "Kitchen lights")
        """

        def _thread_func(_n_tries, errors=None):
            response = None

            if self.action.startswith('procedure.'):
                context['n_tries'] = _n_tries
                response = self._execute_procedure(**context)
                if response is not None:
                    self._send_response(response)
                return response
            # utils.get_context is a special action that simply returns the current context
            elif self.action == 'utils.get_context':
                response = Response(output=context)
                self._send_response(response)
                return response
            else:
                action = self.expand_value_from_context(self.action, **context)
                (module_name, method_name) = get_module_and_method_from_action(action)
                plugin = get_plugin(module_name)

            try:
                # Run the action
                args = self._expand_context(**context)
                args = self.expand_value_from_context(args, **context)
                response = plugin.run(method=method_name, **args)

                if not response:
                    logger.warning('Received null response from action {}'.format(action))
                else:
                    if response.is_error():
                        logger.warning(('Response processed with errors from ' +
                                        'action {}: {}').format(
                            action, str(response)))
                    elif not response.disable_logging:
                        logger.info('Processed response from action {}: {}'.
                                    format(action, str(response)))
            except AssertionError as e:
                plugin.logger.exception(e)
                logger.warning('Assertion error from action [{}]: {}'.format(action, str(e)))
                response = Response(output=None, errors=[str(e)])
            except Exception as e:
                # Retry mechanism
                plugin.logger.exception(e)
                logger.warning(('Uncaught exception while processing response ' +
                                'from action [{}]: {}').format(action, str(e)))

                errors = errors or []
                if str(e) not in errors:
                    errors.append(str(e))

                response = Response(output=None, errors=errors)
                if _n_tries - 1 > 0:
                    logger.info('Reloading plugin {} and retrying'.format(module_name))
                    get_plugin(module_name, reload=True)
                    response = _thread_func(_n_tries=_n_tries-1, errors=errors)
            finally:
                self._send_response(response)
                return response

        token_hash = Config.get('token_hash')

        if token_hash:
            if self.token is None or get_hash(self.token) != token_hash:
                raise PermissionError()

        if _async:
            Thread(target=_thread_func, args=(n_tries,)).start()
        else:
            return _thread_func(n_tries)

    def __str__(self):
        """
        Overrides the str() operator and converts
        the message into a UTF-8 JSON string
        """

        return json.dumps({
            'type': 'request',
            'target': self.target,
            'action': self.action,
            'args': self.args,
            'origin': self.origin if hasattr(self, 'origin') else None,
            'id': self.id if hasattr(self, 'id') else None,
            'token': self.token if hasattr(self, 'token') else None,
            '_timestamp': self.timestamp,
        })

# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class SerialDataEvent(Event):
    """
    Event fired when a serial interface (generic USB, Arduino etc.) receives new data
    """

    def __init__(self, data, device=None, *args, **kwargs):
        """
        :param data: Received data
        :type data: object

        :param device: Source device address or name
        :type device: str
        """

        super().__init__(data=data, device=device, *args, **kwargs)


# vim:sw=4:ts=4:et:




from platypush.message.event import Event


class TodoistEvent(Event):
    pass


class NewItemEvent(TodoistEvent):
    """
    Event triggered when a new item is created.
    """

    def __init__(self, item, *args, **kwargs):
        super().__init__(*args, item=item, **kwargs)


class RemovedItemEvent(TodoistEvent):
    """
    Event triggered when a new item is removed.
    """

    def __init__(self, item, *args, **kwargs):
        super().__init__(*args, item=item, **kwargs)


class ModifiedItemEvent(TodoistEvent):
    """
    Event triggered when an item is changed.
    """

    def __init__(self, item, *args, **kwargs):
        super().__init__(*args, item=item, **kwargs)


class CheckedItemEvent(ModifiedItemEvent):
    """
    Event triggered when an item is checked.
    """

    def __init__(self, item, *args, **kwargs):
        super().__init__(*args, item=item, **kwargs)


class ItemContentChangeEvent(ModifiedItemEvent):
    """
    Event triggered when the content of an item changes.
    """

    def __init__(self, item, *args, **kwargs):
        super().__init__(*args, item=item, **kwargs)


class TodoistSyncRequiredEvent(TodoistEvent):
    """
    Event triggered when an event occurs that doesn't fall into the categories above.
    """


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class LatLongUpdateEvent(Event):
    """
    Event triggered upon GPS location update
    """

    def __init__(self, latitude, longitude, altitude=None, *args, **kwargs):
        """
        :param latitude: GPS latitude
        :type latitude: float

        :param longitude: GPS longitude
        :type longitude: float

        :param altitude: GPS altitude
        :type altitude: float
        """

        super().__init__(*args, latitude=latitude, longitude=longitude, altitude=altitude, **kwargs)


# vim:sw=4:ts=4:et:



import time

from platypush.backend import Backend
from platypush.message.event.joystick import JoystickEvent


class JoystickBackend(Backend):
    """
    This backend will listen for events from a joystick device and post a
    JoystickEvent whenever a new event is captured.

    Triggers:

        * :class:`platypush.message.event.joystick.JoystickEvent` when a new joystick event is received

    Requires:

        * **inputs** (``pip install inputs``)
    """

    def __init__(self, device, *args, **kwargs):
        """
        :param device: Path to the joystick device (e.g. `/dev/input/js0`)
        :type device_name: str
        """

        super().__init__(*args, **kwargs)

        self.device = device

    def run(self):
        import inputs

        super().run()
        self.logger.info('Initialized joystick backend on device {}'.format(self.device))

        while not self.should_stop():
            try:
                events = inputs.get_gamepad()
                for event in events:
                    if event.ev_type == 'Key' or event.ev_type == 'Absolute':
                        self.bus.post(JoystickEvent(code=event.code, state=event.state))
            except Exception as e:
                self.logger.exception(e)
                time.sleep(1)


# vim:sw=4:ts=4:et:




from platypush.message.event import Event


class MediaEvent(Event):
    """ Base class for media events """

    def __init__(self, player=None, plugin=None, *args, **kwargs):
        super().__init__(player=player, plugin=plugin, *args, **kwargs)


class MediaPlayRequestEvent(MediaEvent):
    """
    Event triggered when a new media playback request is received
    """

    def __init__(self, player=None, plugin=None, resource=None, title=None, *args, **kwargs):
        """
        :param resource: File name or URI of the played video
        :type resource: str
        """

        super().__init__(*args, player=player, plugin=plugin, resource=resource, title=title, **kwargs)


class MediaPlayEvent(MediaEvent):
    """
    Event triggered when a new media content is played
    """

    def __init__(self, player=None, plugin=None, resource=None, title=None, *args, **kwargs):
        """
        :param resource: File name or URI of the played video
        :type resource: str
        """

        super().__init__(*args, player=player, plugin=plugin, resource=resource, title=title, **kwargs)


class MediaStopEvent(MediaEvent):
    """
    Event triggered when a media is stopped
    """

    def __init__(self, player=None, plugin=None, *args, **kwargs):
        super().__init__(*args, player=player, plugin=plugin, **kwargs)


class MediaPauseEvent(MediaEvent):
    """
    Event triggered when a media playback is paused
    """

    def __init__(self, player=None, plugin=None, *args, **kwargs):
        super().__init__(*args, player=player, plugin=plugin, **kwargs)


class MediaSeekEvent(MediaEvent):
    """
    Event triggered when the time position in the media changes
    """

    def __init__(self, position, player=None, plugin=None, *args, **kwargs):
        super().__init__(*args, player=player, plugin=plugin, position=position, **kwargs)


class MediaVolumeChangedEvent(MediaEvent):
    """
    Event triggered when the media volume changes
    """

    def __init__(self, volume, player=None, plugin=None, *args, **kwargs):
        super().__init__(*args, player=player, plugin=plugin, volume=volume, **kwargs)


class MediaMuteChangedEvent(MediaEvent):
    """
    Event triggered when the media is muted/unmuted
    """

    def __init__(self, mute, player=None, plugin=None, *args, **kwargs):
        super().__init__(*args, player=player, plugin=plugin, mute=mute, **kwargs)


class NewPlayingMediaEvent(MediaEvent):
    """
    Event triggered when a new media source is being played
    """

    def __init__(self, player=None, plugin=None, resource=None, *args, **kwargs):
        """
        :param resource: File name or URI of the played resource
        :type resource: str
        """

        super().__init__(*args, player=player, plugin=plugin, resource=resource, **kwargs)


# vim:sw=4:ts=4:et:



from platypush.plugins import action
from platypush.plugins.http.request import HttpRequestPlugin

class HttpRequestRssPlugin(HttpRequestPlugin):
    """
    Plugin to programmatically retrieve and parse an RSS feed URL.

    Requires:

        * **feedparser** (``pip install feedparser``)
    """

    @action
    def get(self, url):
        import feedparser
        response = super().get(url, output='text').output
        feed = feedparser.parse(response)
        return feed.entries


# vim:sw=4:ts=4:et:




from platypush.message.event import Event


class SnapcastEvent(Event):
    """ Base class for Snapcast events """

    def __init__(self, host='localhost', *args, **kwargs):
        super().__init__(host=host, *args, **kwargs)


class ClientConnectedEvent(SnapcastEvent):
    """
    Event fired upon client connection
    """

    def __init__(self, client, host='localhost', *args, **kwargs):
        super().__init__(client=client, host=host, *args, **kwargs)


class ClientDisconnectedEvent(SnapcastEvent):
    """
    Event fired upon client disconnection
    """

    def __init__(self, client, host='localhost', *args, **kwargs):
        super().__init__(client=client, host=host, *args, **kwargs)


class ClientVolumeChangeEvent(SnapcastEvent):
    """
    Event fired upon volume change or mute status change on a client
    """

    def __init__(self, client, volume, muted, host='localhost', *args, **kwargs):
        super().__init__(client=client, host=host, volume=volume,
                         muted=muted, *args, **kwargs)


class ClientLatencyChangeEvent(SnapcastEvent):
    """
    Event fired upon latency change on a client
    """

    def __init__(self, client, latency, host='localhost', *args, **kwargs):
        super().__init__(client=client, host=host, latency=latency,
                         *args, **kwargs)


class ClientNameChangeEvent(SnapcastEvent):
    """
    Event fired upon name change of a client
    """

    def __init__(self, client, name, host='localhost', *args, **kwargs):
        super().__init__(client=client, host=host, name=name,
                         *args, **kwargs)


class GroupMuteChangeEvent(SnapcastEvent):
    """
    Event fired upon mute status change
    """

    def __init__(self, group, muted, host='localhost', *args, **kwargs):
        super().__init__(group=group, host=host, muted=muted, *args, **kwargs)


class GroupStreamChangeEvent(SnapcastEvent):
    """
    Event fired upon group stream change
    """

    def __init__(self, group, stream, host='localhost', *args, **kwargs):
        super().__init__(group=group, host=host, stream=stream, *args, **kwargs)


class StreamUpdateEvent(SnapcastEvent):
    """
    Event fired upon stream update
    """

    def __init__(self, stream_id, stream, host='localhost', *args, **kwargs):
        super().__init__(stream_id=stream_id, stream=stream, host=host, *args, **kwargs)


class ServerUpdateEvent(SnapcastEvent):
    """
    Event fired upon stream update
    """

    def __init__(self, server, host='localhost', *args, **kwargs):
        super().__init__(server=server, host=host, *args, **kwargs)


# vim:sw=4:ts=4:et:



import inspect
import os
import platform
import struct
import threading
from typing import Optional

from platypush.message.event.stt import SpeechStartedEvent

from platypush.context import get_bus
from platypush.message.response.stt import SpeechDetectedResponse
from platypush.plugins import action
from platypush.plugins.stt import SttPlugin


class SttPicovoiceSpeechPlugin(SttPlugin):
    """
    This plugin performs speech detection using `PicoVoice <https://github.com/Picovoice>`_.
    NOTE: The PicoVoice product used for real-time speech-to-text (Cheetah) can be used freely for
    personal applications on x86_64 Linux. Other architectures and operating systems require a commercial license.
    You can ask for a license `here <https://picovoice.ai/contact.html>`_.

    Requires:

        * **cheetah** (``pip install git+https://github.com/BlackLight/cheetah``)

    """

    def __init__(self,
                 library_path: Optional[str] = None,
                 acoustic_model_path: Optional[str] = None,
                 language_model_path: Optional[str] = None,
                 license_path: Optional[str] = None,
                 end_of_speech_timeout: int = 1,
                 *args, **kwargs):
        """
        :param library_path: Path to the Cheetah binary library for your OS
            (default: ``CHEETAH_INSTALL_DIR/lib/OS/ARCH/libpv_cheetah.EXT``).
        :param acoustic_model_path: Path to the acoustic speech model
            (default: ``CHEETAH_INSTALL_DIR/lib/common/acoustic_model.pv``).
        :param language_model_path:  Path to the language model
            (default: ``CHEETAH_INSTALL_DIR/lib/common/language_model.pv``).
        :param license_path: Path to your PicoVoice license
            (default: ``CHEETAH_INSTALL_DIR/resources/license/cheetah_eval_linux_public.lic``).
        :param end_of_speech_timeout: Number of seconds of silence during speech recognition before considering
            a phrase over (default: 1).
        """
        from pvcheetah import Cheetah
        super().__init__(*args, **kwargs)

        self._basedir = os.path.abspath(os.path.join(inspect.getfile(Cheetah), '..', '..', '..'))
        if not library_path:
            library_path = self._get_library_path()
        if not language_model_path:
            language_model_path = os.path.join(self._basedir, 'lib', 'common', 'language_model.pv')
        if not acoustic_model_path:
            acoustic_model_path = os.path.join(self._basedir, 'lib', 'common', 'acoustic_model.pv')
        if not license_path:
            license_path = os.path.join(self._basedir, 'resources', 'license', 'cheetah_eval_linux_public.lic')

        self._library_path = library_path
        self._language_model_path = language_model_path
        self._acoustic_model_path = acoustic_model_path
        self._license_path = license_path
        self._end_of_speech_timeout = end_of_speech_timeout
        self._stt_engine: Optional[Cheetah] = None
        self._speech_in_progress = threading.Event()

    def _get_library_path(self) -> str:
        path = os.path.join(self._basedir, 'lib', platform.system().lower(), platform.machine())
        return os.path.join(path, [f for f in os.listdir(path) if f.startswith('libpv_cheetah.')][0])

    def convert_frames(self, frames: bytes) -> tuple:
        assert self._stt_engine, 'The speech engine is not running'
        return struct.unpack_from("h" * self._stt_engine.frame_length, frames)

    def on_detection_ended(self) -> None:
        if self._stt_engine:
            self._stt_engine.delete()
        self._stt_engine = None

    def detect_speech(self, frames: tuple) -> str:
        text, is_endpoint = self._stt_engine.process(frames)
        text = text.strip()

        if text:
            if not self._speech_in_progress.is_set():
                self._speech_in_progress.set()
                get_bus().post(SpeechStartedEvent())

            self._current_text += ' ' + text.strip()

        if is_endpoint:
            text = self._stt_engine.flush().strip().strip()
            if text:
                self._current_text += ' ' + text

            self._speech_in_progress.clear()
            if self._current_text:
                self.on_speech_detected(self._current_text)

            self._current_text = ''

        return self._current_text

    def process_text(self, text: str) -> None:
        pass

    @action
    def detect(self, audio_file: str) -> SpeechDetectedResponse:
        """
        Perform speech-to-text analysis on an audio file.

        :param audio_file: Path to the audio file.
        """
        pass

    def recording_thread(self, input_device: Optional[str] = None, *args, **kwargs) -> None:
        assert self._stt_engine, 'The hotword engine has not yet been initialized'
        super().recording_thread(block_size=self._stt_engine.frame_length, input_device=input_device)

    @action
    def start_detection(self, *args, **kwargs) -> None:
        from pvcheetah import Cheetah
        self._stt_engine = Cheetah(
            library_path=self._library_path,
            acoustic_model_path=self._acoustic_model_path,
            language_model_path=self._language_model_path,
            license_path=self._license_path,
            endpoint_duration_sec=self._end_of_speech_timeout,
        )

        self.rate = self._stt_engine.sample_rate
        self._speech_in_progress.clear()
        super().start_detection(*args, **kwargs)


# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.message.response import Response


class PingResponse(Response):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 host: str,
                 success: bool,
                 *args,
                 min: Optional[float] = None,
                 max: Optional[float] = None,
                 avg: Optional[float] = None,
                 mdev: Optional[float] = None,
                 **kwargs):
        super().__init__(*args, output={
            'host': host,
            'success': success,
            'min': min,
            'max': max,
            'avg': avg,
            'mdev': mdev,
        }, **kwargs)


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class PathOpenEvent(Event):
    """
    Event triggered when a monitored file is opened
    """

    def __init__(self, path, *args, **kwargs):
        """
        :param path: File name
        :type path: str
        """

        super().__init__(path=path, *args, **kwargs)

class PathCloseEvent(Event):
    """
    Event triggered when a monitored file is closed
    """

    def __init__(self, path, *args, **kwargs):
        """
        :param path: File name
        :type path: str
        """

        super().__init__(path=path, *args, **kwargs)

class PathCreateEvent(Event):
    """
    Event triggered when a monitored file is created
    """

    def __init__(self, path, *args, **kwargs):
        """
        :param path: File name
        :type path: str
        """

        super().__init__(path=path, *args, **kwargs)

class PathDeleteEvent(Event):
    """
    Event triggered when a monitored file is deleted
    """

    def __init__(self, path, *args, **kwargs):
        """
        :param path: File name
        :type path: str
        """

        super().__init__(path=path, *args, **kwargs)

class PathModifyEvent(Event):
    """
    Event triggered when a monitored file is modified
    """

    def __init__(self, path, *args, **kwargs):
        """
        :param path: File name
        :type path: str
        """

        super().__init__(path=path, *args, **kwargs)

class PathPermissionsChangeEvent(Event):
    """
    Event triggered when the permissions on a monitored file are changed
    """

    def __init__(self, path, umask, *args, **kwargs):
        """
        :param path: File name
        :type path: str

        :param umask: New file umask
        :type umask: int
        """

        super().__init__(path=path, umask=umask, *args, **kwargs)


# vim:sw=4:ts=4:et:




"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import json
import requests

from platypush.plugins import action
from platypush.plugins.google import GooglePlugin


class GoogleMapsPlugin(GooglePlugin):
    """
    Plugins that provides utilities to interact with Google Maps API services.
    """

    scopes = []

    def __init__(self, api_key, *args, **kwargs):
        """
        :param api_key: Server-side API key to be used for the requests, get one at https://console.developers.google.com
        :type api_key: str
        """

        super().__init__(scopes=self.scopes, *args, **kwargs)
        self.api_key = api_key


    @action
    def get_address_from_latlng(self, latitude, longitude):
        """
        Get an address information given lat/long

        :param latitude: Latitude
        :type latitude: float

        :param longitude: Longitude
        :type longitude: float
        """

        response = requests.get('https://maps.googleapis.com/maps/api/geocode/json',
                                params = {
                                    'latlng': '{},{}'.format(latitude, longitude),
                                    'key': self.api_key,
                                }).json()

        address = dict(
            (t, None) for t in ['street_number', 'street', 'locality', 'country', 'postal_code']
        )

        address['latitude'] = latitude
        address['longitude'] = longitude

        if 'results' in response and response['results']:
            result = response['results'][0]
            self.logger.info('Google Maps geocode response for latlng ({},{}): {}'.
                         format(latitude, longitude, result))

            address['address'] = result['formatted_address'].split(',')[0]
            for addr_component in result['address_components']:
                for component_type in addr_component['types']:
                    if component_type == 'locality':
                        address['locality'] = addr_component['long_name']
                    elif component_type == 'country':
                        address['country'] = addr_component['short_name'].lower()
                    elif component_type == 'postal_code':
                        address['postal_code'] = addr_component['long_name']

        return address

    @action
    def get_elevation_from_latlng(self, latitude, longitude):
        """
        Get the elevation in meters of a geo point given lat/long

        :param latitude: Latitude
        :type latitude: float

        :param longitude: Longitude
        :type longitude: float
        """

        response = requests.get('https://maps.googleapis.com/maps/api/elevation/json',
                                params = {
                                    'locations': '{},{}'.format(latitude, longitude),
                                    'key': self.api_key,
                                }).json()

        elevation = None

        if response.get('results'):
            elevation = response['results'][0]['elevation']

        return { 'elevation': elevation }


# vim:sw=4:ts=4:et:




import re

from typing import Type, Optional, Union, List

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.chat.telegram import MessageEvent, CommandMessageEvent, TextMessageEvent, \
    PhotoMessageEvent, VideoMessageEvent, ContactMessageEvent, DocumentMessageEvent, LocationMessageEvent, \
    GroupChatCreatedEvent
from platypush.plugins.chat.telegram import ChatTelegramPlugin


class ChatTelegramBackend(Backend):
    """
    Telegram bot that listens for messages and updates.

    Triggers:

        * :class:`platypush.message.event.chat.telegram.TextMessageEvent` when a text message is received.
        * :class:`platypush.message.event.chat.telegram.PhotoMessageEvent` when a photo is received.
        * :class:`platypush.message.event.chat.telegram.VideoMessageEvent` when a video is received.
        * :class:`platypush.message.event.chat.telegram.LocationMessageEvent` when a location is received.
        * :class:`platypush.message.event.chat.telegram.ContactMessageEvent` when a contact is received.
        * :class:`platypush.message.event.chat.telegram.DocumentMessageEvent` when a document is received.
        * :class:`platypush.message.event.chat.telegram.CommandMessageEvent` when a command message is received.
        * :class:`platypush.message.event.chat.telegram.GroupCreatedEvent` when the bot is invited to a new group.

    Requires:

        * The :class:`platypush.plugins.chat.telegram.ChatTelegramPlugin` plugin configured

    """

    def __init__(self, authorized_chat_ids: Optional[List[Union[str, int]]] = None, **kwargs):
        """
        :param authorized_chat_ids: Optional list of chat_id/user_id which are authorized to send messages to
            the bot. If nothing is specified then no restrictions are applied.
        """

        super().__init__(**kwargs)
        self.authorized_chat_ids = set(authorized_chat_ids or [])
        self._plugin: ChatTelegramPlugin = get_plugin('chat.telegram')

    def _authorize(self, msg):
        if not self.authorized_chat_ids:
            return

        if msg.chat.type == 'private' and msg.chat.id not in self.authorized_chat_ids:
            self.logger.info('Received message from unauthorized chat_id {}'.format(msg.chat.id))
            self._plugin.send_message(chat_id=msg.chat.id, text='You are not allowed to send messages to this bot')
            raise PermissionError

    def _msg_hook(self, cls: Type[MessageEvent]):
        # noinspection PyUnusedLocal
        def hook(update, context):
            msg = update.effective_message

            try:
                self._authorize(msg)
                self.bus.post(cls(chat_id=update.effective_chat.id,
                                  message=self._plugin.parse_msg(msg).output,
                                  user=self._plugin.parse_user(update.effective_user).output))
            except PermissionError:
                pass

        return hook

    def _group_hook(self):
        # noinspection PyUnusedLocal
        def hook(update, context):
            msg = update.effective_message
            if msg.group_chat_created:
                self.bus.post(GroupChatCreatedEvent(chat_id=update.effective_chat.id,
                                                    message=self._plugin.parse_msg(msg).output,
                                                    user=self._plugin.parse_user(update.effective_user).output))
            elif msg.photo:
                self._msg_hook(PhotoMessageEvent)(update, context)
            elif msg.video:
                self._msg_hook(VideoMessageEvent)(update, context)
            elif msg.contact:
                self._msg_hook(ContactMessageEvent)(update, context)
            elif msg.location:
                self._msg_hook(LocationMessageEvent)(update, context)
            elif msg.document:
                self._msg_hook(DocumentMessageEvent)(update, context)
            elif msg.text:
                if msg.text.startswith('/'):
                    self._command_hook()(update, context)
                else:
                    self._msg_hook(TextMessageEvent)(update, context)

        return hook

    def _command_hook(self):
        # noinspection PyUnusedLocal
        def hook(update, context):
            msg = update.effective_message
            m = re.match('\s*/([0-9a-zA-Z_-]+)\s*(.*)', msg.text)
            cmd = m.group(1).lower()
            args = [arg for arg in re.split('\s+', m.group(2)) if len(arg)]

            try:
                self._authorize(msg)
                self.bus.post(CommandMessageEvent(chat_id=update.effective_chat.id,
                                                  command=cmd,
                                                  cmdargs=args,
                                                  message=self._plugin.parse_msg(msg).output,
                                                  user=self._plugin.parse_user(update.effective_user).output))
            except PermissionError:
                pass

        return hook

    def run(self):
        # noinspection PyPackageRequirements
        from telegram.ext import MessageHandler, Filters

        super().run()
        telegram = self._plugin.get_telegram()
        dispatcher = telegram.dispatcher

        dispatcher.add_handler(MessageHandler(Filters.group, self._group_hook()))
        dispatcher.add_handler(MessageHandler(Filters.text, self._msg_hook(TextMessageEvent)))
        dispatcher.add_handler(MessageHandler(Filters.photo, self._msg_hook(PhotoMessageEvent)))
        dispatcher.add_handler(MessageHandler(Filters.video, self._msg_hook(VideoMessageEvent)))
        dispatcher.add_handler(MessageHandler(Filters.contact, self._msg_hook(ContactMessageEvent)))
        dispatcher.add_handler(MessageHandler(Filters.location, self._msg_hook(LocationMessageEvent)))
        dispatcher.add_handler(MessageHandler(Filters.document, self._msg_hook(DocumentMessageEvent)))
        dispatcher.add_handler(MessageHandler(Filters.command, self._command_hook()))

        self.logger.info('Initialized Telegram backend')
        telegram.start_polling()


# vim:sw=4:ts=4:et:



import os

from flask import Flask

from platypush.backend.http.app.utils import get_routes


## Webapp initialization

base_folder = os.path.abspath(os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '..'))

template_folder = os.path.join(base_folder, 'templates')
static_folder = os.path.join(base_folder, 'static')

application = Flask('platypush', template_folder=template_folder,
                    static_folder=static_folder)

for route in get_routes():
    application.register_blueprint(route)


# vim:sw=4:ts=4:et:



import os
from typing import Optional, Union

import deepspeech
import numpy as np
import wave

from platypush.message.response.stt import SpeechDetectedResponse
from platypush.plugins import action
from platypush.plugins.stt import SttPlugin


class SttDeepspeechPlugin(SttPlugin):
    """
    This plugin performs speech-to-text and speech detection using the
    `Mozilla DeepSpeech <https://github.com/mozilla/DeepSpeech>`_ engine.

    Requires:

        * **deepspeech** (``pip install 'deepspeech>=0.6.0'``)
        * **numpy** (``pip install numpy``)
        * **sounddevice** (``pip install sounddevice``)

    """

    def __init__(self,
                 model_file: str,
                 lm_file: str,
                 trie_file: str,
                 lm_alpha: float = 0.75,
                 lm_beta: float = 1.85,
                 beam_width: int = 500,
                 *args, **kwargs):
        """
        In order to run the speech-to-text engine you'll need to download the right model files for the
        Deepspeech engine that you have installed:

        .. code-block:: shell

            # Create the working folder for the models
            export MODELS_DIR=~/models
            mkdir -p $MODELS_DIR
            cd $MODELS_DIR

            # Download and extract the model files for your version of Deepspeech. This may take a while.
            export DEEPSPEECH_VERSION=0.6.1
            wget https://github.com/mozilla/DeepSpeech/releases/download/v$DEEPSPEECH_VERSION/deepspeech-$DEEPSPEECH_VERSION-models.tar.gz
            tar -xvzf deepspeech-$DEEPSPEECH_VERSION-models.tar.gz
            x deepspeech-0.6.1-models/
            x deepspeech-0.6.1-models/lm.binary
            x deepspeech-0.6.1-models/output_graph.pbmm
            x deepspeech-0.6.1-models/output_graph.pb
            x deepspeech-0.6.1-models/trie
            x deepspeech-0.6.1-models/output_graph.tflite

        :param model_file: Path to the model file (usually named ``output_graph.pb`` or ``output_graph.pbmm``).
            Note that ``.pbmm`` usually perform better and are smaller.

        :param lm_file: Path to the language model binary file (usually named ``lm.binary``).
        :param trie_file: The path to the trie file build from the same vocabulary as the language model binary
            (usually named ``trie``).
        :param lm_alpha: The alpha hyperparameter of the CTC decoder - Language Model weight.
            See <https://github.com/mozilla/DeepSpeech/releases/tag/v0.6.0>.
        :param lm_beta: The beta hyperparameter of the CTC decoder - Word Insertion weight.
            See <https://github.com/mozilla/DeepSpeech/releases/tag/v0.6.0>.
        :param beam_width:  Decoder beam width (see beam scoring in KenLM language model).
        :param input_device: PortAudio device index or name that will be used for recording speech (default: default
            system audio input device).
        :param hotword: When this word is detected, the plugin will trigger a
            :class:`platypush.message.event.stt.HotwordDetectedEvent` instead of a
            :class:`platypush.message.event.stt.SpeechDetectedEvent` event. You can use these events for hooking other
            assistants.
        :param hotwords: Use a list of hotwords instead of a single one.
        :param conversation_timeout: If ``hotword`` or ``hotwords`` are set and ``conversation_timeout`` is set,
            the next speech detected event will trigger a :class:`platypush.message.event.stt.ConversationDetectedEvent`
            instead of a :class:`platypush.message.event.stt.SpeechDetectedEvent` event. You can hook custom hooks
            here to run any logic depending on the detected speech - it can emulate a kind of
            "OK, Google. Turn on the lights" interaction without using an external assistant.
        :param block_duration: Duration of the acquired audio blocks (default: 1 second).
        """

        super().__init__(*args, **kwargs)
        self.model_file = os.path.abspath(os.path.expanduser(model_file))
        self.lm_file = os.path.abspath(os.path.expanduser(lm_file))
        self.trie_file = os.path.abspath(os.path.expanduser(trie_file))
        self.lm_alpha = lm_alpha
        self.lm_beta = lm_beta
        self.beam_width = beam_width
        self._model: Optional[deepspeech.Model] = None
        self._context = None

    def _get_model(self) -> deepspeech.Model:
        if not self._model:
            self._model = deepspeech.Model(self.model_file, self.beam_width)
            self._model.enableDecoderWithLM(self.lm_file, self.trie_file, self.lm_alpha, self.lm_beta)

        return self._model

    def _get_context(self):
        if not self._model:
            self._model = self._get_model()
        if not self._context:
            self._context = self._model.createStream()

        return self._context

    @staticmethod
    def convert_frames(frames: Union[np.ndarray, bytes]) -> np.ndarray:
        return np.frombuffer(frames, dtype=np.int16)

    def on_detection_started(self):
        self._context = self._get_context()

    def on_detection_ended(self):
        if self._model and self._context:
            self._model.finishStream()
        self._context = None

    def detect_speech(self, frames) -> str:
        model = self._get_model()
        context = self._get_context()
        model.feedAudioContent(context, frames)
        return model.intermediateDecode(context)

    def on_speech_detected(self, speech: str) -> None:
        super().on_speech_detected(speech)
        if not speech:
            return

        model = self._get_model()
        context = self._get_context()
        model.finishStream(context)
        self._context = None

    @action
    def detect(self, audio_file: str) -> SpeechDetectedResponse:
        """
        Perform speech-to-text analysis on an audio file.

        :param audio_file: Path to the audio file.
        """
        audio_file = os.path.abspath(os.path.expanduser(audio_file))
        wav = wave.open(audio_file, 'r')
        buffer = wav.readframes(wav.getnframes())
        data = self.convert_frames(buffer)
        model = self._get_model()
        speech = model.stt(data)
        return SpeechDetectedResponse(speech=speech)


# vim:sw=4:ts=4:et:



import datetime
import enum
import os

from sqlalchemy import create_engine, Column, Integer, String, DateTime, \
    Enum, ForeignKey

from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql.expression import func

from platypush.backend.http.request import HttpRequest
from platypush.config import Config
from platypush.context import get_plugin
from platypush.message.event.http.rss import NewFeedEvent

Base = declarative_base()
Session = scoped_session(sessionmaker())


class RssUpdates(HttpRequest):
    """
    Gets new items in an RSS feed

    Requires:

        * **feedparser** (``pip install feedparser``)
    """

    user_agent = 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) ' + \
                 'Chrome/62.0.3202.94 Safari/537.36'

    def __init__(self, url, title=None, headers=None, params=None, max_entries=None,
                 extract_content=False, digest_format=None, *argv, **kwargs):
        self.workdir = os.path.join(os.path.expanduser(Config.get('workdir')), 'feeds')
        self.dbfile = os.path.join(self.workdir, 'rss.db')
        self.url = url
        self.title = title
        self.max_entries = max_entries

        # If true, then the http.webpage plugin will be used to parse the content
        self.extract_content = extract_content

        self.digest_format = digest_format.lower() if digest_format else None  # Supported formats: html, pdf

        os.makedirs(os.path.expanduser(os.path.dirname(self.dbfile)), exist_ok=True)

        if headers is None:
            headers = {}
        headers['User-Agent'] = self.user_agent

        request_args = {
            'method': 'get',
            'url': self.url,
            'headers': headers,
            'params': params or {},
        }

        super().__init__(skip_first_call=False, args=request_args, *argv, **kwargs)

    def _get_or_create_source(self, session):
        record = session.query(FeedSource).filter_by(url=self.url).first()
        if record is None:
            record = FeedSource(url=self.url, title=self.title)
            session.add(record)

        session.commit()
        return record

    @staticmethod
    def _get_latest_update(session, source_id):
        return session.query(func.max(FeedEntry.published)).filter_by(source_id=source_id).scalar()

    def _parse_entry_content(self, link):
        self.logger.info('Extracting content from {}'.format(link))

        parser = get_plugin('http.webpage')
        response = parser.simplify(link).output
        errors = parser.simplify(link).errors

        if not response:
            self.logger.warning('Mercury parser error: '.format(errors or '[unknown error]'))
            return

        return response.get('content')

    def get_new_items(self, response):
        import feedparser
        engine = create_engine('sqlite:///{}'.format(self.dbfile),
                               connect_args={'check_same_thread': False})

        Base.metadata.create_all(engine)
        Session.configure(bind=engine)
        self._get_or_create_source(session=Session())

        feed = feedparser.parse(response.text)
        session = Session()
        source_record = self._get_or_create_source(session=session)
        session.add(source_record)
        parse_start_time = datetime.datetime.utcnow()
        entries = []
        latest_update = self._get_latest_update(session, source_record.id)

        if not self.title and 'title' in feed.feed:
            self.title = feed.feed['title']
            source_record.title = self.title

        content = u'''
            <h1 style="margin-top: 30px">{}</h1>
            <h2 style="margin-top: 10px; page-break-after: always">
                Feeds digest generated on {} </h2>'''.format(self.title,
                                                             datetime.datetime.now().strftime('%d %B %Y, %H:%M'))

        style = '''
            body {
                font-size: 22px;
                font-family: 'Merriweather', Georgia, 'Times New Roman', Times, serif;
            }
            '''

        self.logger.info('Parsed {:d} items from RSS feed <{}>'
                         .format(len(feed.entries), self.url))

        for entry in feed.entries:
            if not entry.published_parsed:
                continue

            try:
                entry_timestamp = datetime.datetime(*entry.published_parsed[:6])

                if latest_update is None \
                        or entry_timestamp > latest_update:
                    self.logger.info('Processed new item from RSS feed <{}>'.format(self.url))
                    entry.summary = entry.summary if hasattr(entry, 'summary') else None

                    if self.extract_content:
                        entry.content = self._parse_entry_content(entry.link)
                    elif hasattr(entry, 'summary'):
                        entry.content = entry.summary
                    else:
                        entry.content = None

                    content += u'''<h1 style="page-break-before: always">{}</h1>
                        <div class="_parsed-content">{}</div>'''.format(entry.title, entry.content)

                    e = {
                        'entry_id': entry.id,
                        'title': entry.title,
                        'link': entry.link,
                        'summary': entry.summary,
                        'content': entry.content,
                        'source_id': source_record.id,
                        'published': entry_timestamp,
                    }

                    entries.append(e)
                    session.add(FeedEntry(**e))
                    if self.max_entries and len(entries) > self.max_entries:
                        break
            except Exception as e:
                self.logger.warning('Exception encountered while parsing RSS ' +
                                    'RSS feed {}: {}'.format(entry.link, str(e)))
                self.logger.exception(e)

        source_record.last_updated_at = parse_start_time
        digest_filename = None

        if entries:
            self.logger.info('Parsed {} new entries from the RSS feed {}'.format(
                len(entries), self.title))

            if self.digest_format:
                digest_filename = os.path.join(self.workdir, 'cache', '{}_{}.{}'.format(
                    datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S'),
                    self.title, self.digest_format))

                os.makedirs(os.path.dirname(digest_filename), exist_ok=True)

                if self.digest_format == 'html':
                    content = '''
                        <html>
                            <head>
                                <title>{title}</title>
                                <style>{style}</style>
                            </head>
                            <body>{{content}}</body>
                        </html>
                    '''.format(title=self.title, style=style, content=content)

                    with open(digest_filename, 'w', encoding='utf-8') as f:
                        f.write(content)
                elif self.digest_format == 'pdf':
                    import weasyprint
                    from weasyprint.fonts import FontConfiguration

                    font_config = FontConfiguration()
                    css = [weasyprint.CSS('https://fonts.googleapis.com/css?family=Merriweather'),
                           weasyprint.CSS(string=style, font_config=font_config)]

                    weasyprint.HTML(string=content).write_pdf(digest_filename, stylesheets=css)
                else:
                    raise RuntimeError('Unsupported format: {}. Supported formats: ' +
                                       'html or pdf'.format(self.digest_format))

                digest_entry = FeedDigest(source_id=source_record.id,
                                          format=self.digest_format,
                                          filename=digest_filename)

                session.add(digest_entry)
                self.logger.info('{} digest ready: {}'.format(self.digest_format, digest_filename))

        session.commit()
        self.logger.info('Parsing RSS feed {}: completed'.format(self.title))

        return NewFeedEvent(request=dict(self), response=entries,
                            source_id=source_record.id, title=self.title,
                            digest_format=self.digest_format,
                            digest_filename=digest_filename)


class FeedSource(Base):
    """ Models the FeedSource table, containing RSS sources to be parsed """

    __tablename__ = 'FeedSource'
    __table_args__ = ({'sqlite_autoincrement': True})

    id = Column(Integer, primary_key=True)
    title = Column(String)
    url = Column(String, unique=True)
    last_updated_at = Column(DateTime)


class FeedEntry(Base):
    """ Models the FeedEntry table, which contains RSS entries """

    __tablename__ = 'FeedEntry'
    __table_args__ = ({'sqlite_autoincrement': True})

    id = Column(Integer, primary_key=True)
    entry_id = Column(String)
    source_id = Column(Integer, ForeignKey('FeedSource.id'), nullable=False)
    title = Column(String)
    link = Column(String)
    summary = Column(String)
    content = Column(String)
    published = Column(DateTime)


class FeedDigest(Base):
    """ Models the FeedDigest table, containing feed digests either in HTML
        or PDF format """

    class DigestFormat(enum.Enum):
        html = 1
        pdf = 2

    __tablename__ = 'FeedDigest'
    __table_args__ = ({'sqlite_autoincrement': True})

    id = Column(Integer, primary_key=True)
    source_id = Column(Integer, ForeignKey('FeedSource.id'), nullable=False)
    format = Column(Enum(DigestFormat), nullable=False)
    filename = Column(String, nullable=False)
    created_at = Column(DateTime, nullable=False, default=datetime.datetime.utcnow)

# vim:sw=4:ts=4:et:



import datetime

from typing import List

from platypush.message import Mapping
from platypush.message.response import Response


class BuienradarWeatherResponse(Response):
    def __init__(self,
                 barometer_fc: str,
                 condition_name: str,
                 condition_name_long: str,
                 condition_image: str,
                 feel_temperature: float,
                 ground_temperature: float,
                 humidity: int,
                 irradiance: int,
                 measured: datetime.datetime,
                 precipitation: float,
                 pressure: float,
                 rain_last_24_hours: float,
                 rain_last_hour: float,
                 station_name: str,
                 temperature: float,
                 visibility: int,
                 wind_azimuth: int,
                 wind_direction: str,
                 wind_force: int,
                 wind_gust: float,
                 wind_speed: float,
                 *args, **kwargs):
        super().__init__(*args, output={
            'barometer_fc': barometer_fc,
            'condition_name': condition_name,
            'condition_name_long': condition_name_long,
            'condition_image': condition_image,
            'feel_temperature': feel_temperature,
            'ground_temperature': ground_temperature,
            'humidity': humidity,
            'irradiance': irradiance,
            'measured': measured,
            'precipitation': precipitation,
            'pressure': pressure,
            'rain_last_24_hours': rain_last_24_hours,
            'rain_last_hour': rain_last_hour,
            'station_name': station_name,
            'temperature': temperature,
            'visibility': visibility,
            'wind_azimuth': wind_azimuth,
            'wind_direction': wind_direction,
            'wind_force': wind_force,
            'wind_gust': wind_gust,
            'wind_speed': wind_speed,
        }, **kwargs)


class BuienradarPrecipitationResponse(Response):
    def __init__(self,
                 average: float,
                 total: float,
                 time_frame: int,
                 *args, **kwargs):
        super().__init__(*args, output={
            'average': average,
            'total': total,
            'time_frame': time_frame,
        }, **kwargs)


class BuienradarForecast(Mapping):
    def __init__(self,
                 condition_name: str,
                 condition_name_long: str,
                 condition_image: str,
                 date_time: datetime.datetime,
                 rain: float,
                 min_rain: float,
                 max_rain: float,
                 rain_chance: float,
                 snow: int,
                 temperature: float,
                 wind_azimuth: int,
                 wind_direction: str,
                 wind_force: int,
                 wind_speed: float,
                 *args, **kwargs):
        super().__init__(*args, output={
            'condition_name': condition_name,
            'condition_name_long': condition_name_long,
            'condition_image': condition_image,
            'date_time': date_time,
            'rain': rain,
            'min_rain': min_rain,
            'max_rain': max_rain,
            'rain_chance': rain_chance,
            'snow': snow,
            'temperature': temperature,
            'wind_azimuth': wind_azimuth,
            'wind_direction': wind_direction,
            'wind_force': wind_force,
            'wind_speed': wind_speed,
        }, **kwargs)


class BuienradarForecastResponse(Response):
    def __init__(self,
                 forecast=List[BuienradarForecast],
                 *args, **kwargs):
        super().__init__(*args, output=forecast, **kwargs)


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class HttpEvent(Event):
    """
    Event triggered upon HTTP request/response cycle completion
    """

    def __init__(self, request, response, *args, **kwargs):
        """
        :param request: Reference to the original HTTP request
        :type request: dict

        :param response: The server response
        :type response: dict or list
        """

        super().__init__(request=request, response=response, *args, **kwargs)


# vim:sw=4:ts=4:et:




import ast
import hashlib
import importlib
import inspect
import logging
import os
import re
import signal
import socket
import ssl
import urllib.request

logger = logging.getLogger(__name__)


def get_module_and_method_from_action(action):
    """ Input  : action=music.mpd.play
        Output : ('music.mpd', 'play') """

    tokens = action.split('.')
    module_name = str.join('.', tokens[:-1])
    method_name = tokens[-1:][0]
    return module_name, method_name


def get_message_class_by_type(msgtype):
    """ Gets the class of a message type given as string """

    try:
        module = importlib.import_module('platypush.message.' + msgtype)
    except ImportError as e:
        logger.warning('Unsupported message type {}'.format(msgtype))
        raise RuntimeError(e)

    cls_name = msgtype[0].upper() + msgtype[1:]

    try:
        msgclass = getattr(module, cls_name)
    except AttributeError as e:
        logger.warning('No such class in {}: {}'.format(
            module.__name__, cls_name))
        raise RuntimeError(e)

    return msgclass


def get_event_class_by_type(type):
    """ Gets an event class by type name """
    event_module = importlib.import_module('.'.join(type.split('.')[:-1]))
    return getattr(event_module, type.split('.')[-1])


def get_plugin_module_by_name(plugin_name):
    """ Gets the module of a plugin by name (e.g. "music.mpd" or "media.vlc") """

    module_name = 'platypush.plugins.' + plugin_name
    try:
        return importlib.import_module('platypush.plugins.' + plugin_name)
    except ImportError as e:
        logger.error('Cannot import {}: {}'.format(module_name, str(e)))
        return None


def get_plugin_class_by_name(plugin_name):
    """ Gets the class of a plugin by name (e.g. "music.mpd" or "media.vlc") """

    module = get_plugin_module_by_name(plugin_name)
    if not module:
        return

    class_name = getattr(module, ''.join([_.capitalize() for _ in plugin_name.split('.')]) + 'Plugin')
    try:
        return getattr(module, ''.join([_.capitalize() for _ in plugin_name.split('.')]) + 'Plugin')
    except Exception as e:
        logger.error('Cannot import class {}: {}'.format(class_name, str(e)))
        return None


def set_timeout(seconds, on_timeout):
    """
    Set a function to be called if timeout expires without being cleared.
    It only works on the main thread.

    Params:
        seconds    -- Timeout in seconds
        on_timeout -- Function invoked on timeout unless clear_timeout is called before
    """

    def _sighandler(signum, frame):
        on_timeout()

    signal.signal(signal.SIGALRM, _sighandler)
    signal.alarm(seconds)


def clear_timeout():
    """ Clear any previously set timeout """
    signal.alarm(0)


def get_hash(s):
    """ Get the SHA256 hash hexdigest of a string input """
    return hashlib.sha256(s.encode('utf-8')).hexdigest()


def get_decorators(cls, climb_class_hierarchy=False):
    """
    Get the decorators of a class as a {"decorator_name": [list of methods]} dictionary
    :param climb_class_hierarchy: If set to True (default: False), it will search return the decorators in the parent classes as well
    :type climb_class_hierarchy: bool
    """

    decorators = {}

    def visit_FunctionDef(node):
        for n in node.decorator_list:
            name = ''
            if isinstance(n, ast.Call):
                name = n.func.attr if isinstance(n.func, ast.Attribute) else n.func.id
            else:
                name = n.attr if isinstance(n, ast.Attribute) else n.id

            decorators[name] = decorators.get(name, set())
            decorators[name].add(node.name)

    if climb_class_hierarchy:
        targets = inspect.getmro(cls)
    else:
        targets = [cls]

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_FunctionDef

    for target in targets:
        try:
            node_iter.visit(ast.parse(inspect.getsource(target)))
        except TypeError:
            # Ignore built-in classes
            pass

    return decorators


def get_redis_queue_name_by_message(msg):
    from platypush.message import Message

    if not isinstance(msg, Message):
        logger.warning('Not a valid message (type: {}): {}'.format(type(msg), msg))

    return 'platypush/responses/{}'.format(msg.id) if msg.id else None


def _get_ssl_context(context_type=None, ssl_cert=None, ssl_key=None,
                     ssl_cafile=None, ssl_capath=None):
    if not context_type:
        ssl_context = ssl.create_default_context(cafile=ssl_cafile,
                                                 capath=ssl_capath)
    else:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)

    if ssl_cafile or ssl_capath:
        ssl_context.load_verify_locations(
            cafile=ssl_cafile, capath=ssl_capath)

    ssl_context.load_cert_chain(
        certfile=os.path.abspath(os.path.expanduser(ssl_cert)),
        keyfile=os.path.abspath(os.path.expanduser(ssl_key)) if ssl_key else None
    )

    return ssl_context


def get_ssl_context(ssl_cert=None, ssl_key=None, ssl_cafile=None,
                    ssl_capath=None):
    return _get_ssl_context(context_type=None,
                            ssl_cert=ssl_cert, ssl_key=ssl_key,
                            ssl_cafile=ssl_cafile, ssl_capath=ssl_capath)


def get_ssl_server_context(ssl_cert=None, ssl_key=None, ssl_cafile=None,
                           ssl_capath=None):
    return _get_ssl_context(context_type=ssl.PROTOCOL_TLS_SERVER,
                            ssl_cert=ssl_cert, ssl_key=ssl_key,
                            ssl_cafile=ssl_cafile, ssl_capath=ssl_capath)


def get_ssl_client_context(ssl_cert=None, ssl_key=None, ssl_cafile=None,
                           ssl_capath=None):
    return _get_ssl_context(context_type=ssl.PROTOCOL_TLS_CLIENT,
                            ssl_cert=ssl_cert, ssl_key=ssl_key,
                            ssl_cafile=ssl_cafile, ssl_capath=ssl_capath)

def set_thread_name(name):
    global logger

    try:
        import prctl
        prctl.set_name(name)
    except ImportError:
        logger.debug('Unable to set thread name: prctl module is missing')


def find_bins_in_path(bin_name):
    return [os.path.join(p, bin_name)
            for p in os.environ.get('PATH', '').split(':')
            if os.path.isfile(os.path.join(p, bin_name))
            and (os.name == 'nt' or
                 os.access(os.path.join(p, bin_name), os.X_OK))]


def find_files_by_ext(directory, *exts):
    """
    Finds all the files in the given directory with the provided extensions
    """

    if not exts:
        raise AttributeError('No extensions provided')

    if not os.path.isdir(directory):
        raise AttributeError('{} is not a valid directory'.format(directory))

    min_len = len(min(exts, key=len))
    max_len = len(max(exts, key=len))
    result = []

    for root, dirs, files in os.walk(directory):
        for i in range(min_len, max_len+1):
            result += [f for f in files if f[-i:] in exts]

    return result


def is_process_alive(pid):
    try:
        os.kill(pid, 0)
        return True
    except OSError:
        return False


def get_ip_or_hostname():
    ip = socket.gethostbyname(socket.gethostname())
    return socket.getfqdn() if ip.startswith('127.') else ip


def get_mime_type(resource):
    import magic
    if resource.startswith('file://'):
        resource = resource[len('file://'):]

    if resource.startswith('http://') or resource.startswith('https://'):
        with urllib.request.urlopen(resource) as response:
            return response.info().get_content_type()
    else:
        if hasattr(magic, 'detect_from_filename'):
            mime = magic.detect_from_filename(resource)
        elif hasattr(magic, 'from_file'):
            mime = magic.from_file(resource, mime=True)
        else:
            raise RuntimeError('The installed magic version provides neither detect_from_filename nor from_file')

        if mime:
            return mime.mime_type


def camel_case_to_snake_case(string):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def grouper(n, iterable, fillvalue=None):
    """
    Split an iterable in groups of max N elements.
    grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx
    """
    from itertools import zip_longest
    args = [iter(iterable)] * n

    if fillvalue:
        return zip_longest(fillvalue=fillvalue, *args)

    for chunk in zip_longest(*args):
        yield filter(None, chunk)


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class DashboardIframeUpdateEvent(Event):
    """
    Deliver a DashboardIframeUpdateEvent if you are using the web dashboard
    and you want the connected web clients to show a certain URL in the iframe
    modal window for (optionally) a certain time.
    """

    def __init__(self, url, width=None, height=None, timeout=None, *args, **kwargs):
        """
        :param url: URL to show in the iframe dashboard element
        :type url: str

        :param width: Iframe width, as int (pixels) or CSS string
        :param height: Iframe height, as int (pixels) or CSS string

        :param timeout: If set, the iframe will be closed after this time (in seconds)
        :type timeout: float
        """

        super().__init__(url=url, width=width, height=height, timeout=timeout, *args, **kwargs)


# vim:sw=4:ts=4:et:




from typing import Optional, Dict, Union

from platypush.message.event import Event


class TensorflowEvent(Event):
    def __init__(self, model: str, logs: Optional[Dict[str, Union[int, float]]], *args, **kwargs):
        """
        :param model: Name of the Tensorflow model.
        :param logs: Logs and metrics.
        """
        super().__init__(*args, model=model, logs=logs, **kwargs)


class TensorflowEpochStartedEvent(TensorflowEvent):
    """
    Triggered when a Tensorflow model training/evaluation epoch begins.
    """
    def __init__(self, epoch: int, *args, **kwargs):
        """
        :param epoch: Epoch index.
        """
        super().__init__(*args, epoch=epoch, **kwargs)


class TensorflowEpochEndedEvent(TensorflowEvent):
    """
    Triggered when a Tensorflow model training/evaluation epoch ends.
    """
    def __init__(self, epoch: int, *args, **kwargs):
        """
        :param epoch: Epoch index.
        """
        super().__init__(*args, epoch=epoch, **kwargs)


class TensorflowBatchStartedEvent(TensorflowEvent):
    """
    Triggered when a Tensorflow model training/evaluation batch starts being processed.
    """
    def __init__(self, batch: int, *args, **kwargs):
        """
        :param batch: Batch index.
        """
        super().__init__(*args, batch=batch, **kwargs)


class TensorflowBatchEndedEvent(TensorflowEvent):
    """
    Triggered when a the processing of a Tensorflow model training/evaluation batch ends.
    """
    def __init__(self, batch: int, *args, **kwargs):
        """
        :param batch: Batch index.
        """
        super().__init__(*args, batch=batch, **kwargs)


class TensorflowTrainStartedEvent(TensorflowEvent):
    """
    Triggered when a Tensorflow model starts being trained.
    """


class TensorflowTrainEndedEvent(TensorflowEvent):
    """
    Triggered when the training phase of a Tensorflow model ends.
    """


# vim:sw=4:ts=4:et:



import re
import threading
import time

from platypush.plugins import action
from platypush.plugins.music import MusicPlugin


class MusicMpdPlugin(MusicPlugin):
    """
    This plugin allows you to interact with an MPD/Mopidy music server.  MPD
    (https://www.musicpd.org/) is a flexible server-side protocol/application
    for handling music collections and playing music, mostly aimed to manage
    local libraries. Mopidy (https://www.mopidy.com/) is an evolution of MPD,
    compatible with the original protocol and with support for multiple music
    sources through plugins (e.g. Spotify, TuneIn, Soundcloud, local files
    etc.).

    Requires:

        * **python-mpd2** (``pip install python-mpd2``)
    """

    _client_lock = threading.RLock()

    def __init__(self, host, port=6600):
        """
        :param host: MPD IP/hostname
        :type host: str

        :param port: MPD port (default: 6600)
        :type port: int
        """

        super().__init__()
        self.host = host
        self.port = port
        self.client = None

    def _connect(self, n_tries=2):
        import mpd

        with self._client_lock:
            if self.client:
                return

            error = None
            while n_tries > 0:
                try:
                    n_tries -= 1
                    self.client = mpd.MPDClient(use_unicode=True)
                    self.client.connect(self.host, self.port)
                    return self.client
                except Exception as e:
                    error = e
                    self.logger.warning('Connection exception: {}{}'.
                                        format(str(e), (': Retrying' if n_tries > 0 else '')))
                    time.sleep(0.5)

        self.client = None
        raise error

    def _exec(self, method, *args, **kwargs):
        error = None
        n_tries = int(kwargs.pop('n_tries')) if 'n_tries' in kwargs else 2
        return_status = kwargs.pop('return_status') \
            if 'return_status' in kwargs else True

        while n_tries > 0:
            try:
                self._connect()
                n_tries -= 1
                with self._client_lock:
                    response = getattr(self.client, method)(*args, **kwargs)

                if return_status:
                    return self.status().output
                return response
            except Exception as e:
                error = str(e)
                self.logger.warning('Exception while executing MPD method {}: {}'.
                                    format(method, error))
                self.client = None

        return None, error

    @action
    def play(self, resource=None):
        """
        Play a resource by path/URI

        :param resource: Resource path/URI
        :type resource: str
        """

        if resource:
            self.add(resource, position=0)
            return self.play_pos(0)

        return self._exec('play')

    @action
    def play_pos(self, pos):
        """
        Play a track in the current playlist by position number

        :param pos: Position number
        """

        return self._exec('play', pos)

    @action
    def pause(self):
        """ Pause playback """

        status = self.status().output['state']
        if status == 'play':
            return self._exec('pause')
        else:
            return self._exec('play')

    @action
    def pause_if_playing(self):
        """ Pause playback only if it's playing """

        status = self.status().output['state']
        if status == 'play':
            return self._exec('pause')

    @action
    def play_if_paused(self):
        """ Play only if it's paused (resume) """

        status = self.status().output['state']
        if status == 'pause':
            return self._exec('play')

    @action
    def play_if_paused_or_stopped(self):
        """ Play only if it's paused or stopped """

        status = self.status().output['state']
        if status == 'pause' or status == 'stop':
            return self._exec('play')

    @action
    def stop(self):
        """ Stop playback """
        return self._exec('stop')

    @action
    def play_or_stop(self):
        """ Play or stop (play state toggle) """
        status = self.status().output['state']
        if status == 'play':
            return self._exec('stop')
        else:
            return self._exec('play')

    @action
    def playid(self, track_id):
        """
        Play a track by ID

        :param track_id: Track ID
        :type track_id: str
        """

        return self._exec('playid', track_id)

    @action
    def next(self):
        """ Play the next track """
        return self._exec('next')

    @action
    def previous(self):
        """ Play the previous track """
        return self._exec('previous')

    @action
    def setvol(self, vol):
        """
        Set the volume

        :param vol: Volume value (range: 0-100)
        :type vol: int
        """
        return self._exec('setvol', str(vol))

    @action
    def volup(self, delta=10):
        """
        Turn up the volume

        :param delta: Volume up delta (default: +10%)
        :type delta: int
        """

        volume = int(self.status().output['volume'])
        new_volume = min(volume + delta, 100)
        return self.setvol(new_volume)

    @action
    def voldown(self, delta=10):
        """
        Turn down the volume

        :param delta: Volume down delta (default: -10%)
        :type delta: int
        """

        volume = int(self.status().output['volume'])
        new_volume = max(volume - delta, 0)
        return self.setvol(new_volume)

    @action
    def random(self, value=None):
        """
        Set random mode

        :param value: If set, set the random state this value (true/false). Default: None (toggle current state)
        :type value: bool
        """

        if value is None:
            value = int(self.status().output['random'])
            value = 1 if value == 0 else 0
        return self._exec('random', value)

    @action
    def consume(self, value=None):
        """
        Set consume mode

        :param value: If set, set the consume state this value (true/false). Default: None (toggle current state)
        :type value: bool
        """

        if value is None:
            value = int(self.status().output['consume'])
            value = 1 if value == 0 else 0
        return self._exec('consume', value)

    @action
    def single(self, value=None):
        """
        Set single mode

        :param value: If set, set the consume state this value (true/false). Default: None (toggle current state)
        :type value: bool
        """

        if value is None:
            value = int(self.status().output['single'])
            value = 1 if value == 0 else 0
        return self._exec('single', value)

    @action
    def repeat(self, value=None):
        """
        Set repeat mode

        :param value: If set, set the repeat state this value (true/false). Default: None (toggle current state)
        :type value: bool
        """

        if value is None:
            value = int(self.status().output['repeat'])
            value = 1 if value == 0 else 0
        return self._exec('repeat', value)

    @action
    def shuffle(self):
        """
        Shuffles the current playlist
        """

        return self._exec('shuffle')

    @action
    def save(self, name):
        """
        Save the current tracklist to a new playlist with the specified name

        :param name: Name of the playlist
        :type name: str
        """
        return self._exec('save', name)

    @action
    def add(self, resource, position=None):
        """
        Add a resource (track, album, artist, folder etc.) to the current playlist

        :param resource: Resource path or URI
        :type resource: str

        :param position: Position where the track(s) will be inserted (default: end of the playlist)
        :type position: int
        """

        if isinstance(resource, list):
            for r in resource:
                r = self._parse_resource(r)
                try:
                    if position is None:
                        self._exec('add', r)
                    else:
                        self._exec('addid', r, position)
                except Exception as e:
                    self.logger.warning('Could not add {}: {}'.format(r, e))

            return self.status().output

        r = self._parse_resource(resource)

        if position is None:
            return self._exec('add', r)
        return self._exec('addid', r, position)

    @action
    def delete(self, positions):
        """
        Delete the playlist item(s) in the specified position(s).

        :param positions: Positions of the tracks to be removed
        :type positions: list[int]

        :return: The modified playlist
        """

        for pos in sorted(positions, key=int, reverse=True):
            self._exec('delete', pos)
        return self.playlistinfo()

    @action
    def rm(self, playlist):
        """
        Permanently remove playlist(s) by name

        :param playlist: Name or list of playlist names to remove
        :type playlist: str or list[str]
        """

        if isinstance(playlist, str):
            playlist = [playlist]
        elif not isinstance(playlist, list):
            raise RuntimeError('Invalid type for playlist: {}'.format(type(playlist)))

        for p in playlist:
            self._exec('rm', p)

    @action
    def move(self, from_pos, to_pos):
        """
        Move the playlist item in position <from_pos> to position <to_pos>

        :param from_pos: Track current position
        :type from_pos: int

        :param to_pos: Track new position
        :type to_pos: int
        """
        return self._exec('move', from_pos, to_pos)

    @classmethod
    def _parse_resource(cls, resource):
        if not resource:
            return

        m = re.search('^https://open.spotify.com/([^?]+)', resource)
        if m:
            resource = 'spotify:{}'.format(m.group(1).replace('/', ':'))

        if resource.startswith('spotify:'):
            resource = resource.split('?')[0]

        m = re.match('spotify:playlist:(.*)', resource)
        if m:
            # Old Spotify URI format, convert it to new
            resource = 'spotify:user:spotify:playlist:' + m.group(1)
        return resource

    @action
    def load(self, playlist, play=True):
        """
        Load and play a playlist by name

        :param playlist: Playlist name
        :type playlist: str

        :param play: Start playback after loading the playlist (default: True)
        :type play: bool
        """

        ret = self._exec('load', playlist)
        if play:
            self.play()
        return ret

    @action
    def clear(self):
        """ Clear the current playlist """
        return self._exec('clear')

    @action
    def seekcur(self, value):
        """
        Seek to the specified position

        :param value: Seek position in seconds, or delta string (e.g. '+15' or '-15') to indicate a seek relative to the current position
        :type value: int
        """

        return self._exec('seekcur', value)

    @action
    def forward(self):
        """ Go forward by 15 seconds """

        return self._exec('seekcur', '+15')

    @action
    def back(self):
        """ Go backward by 15 seconds """

        return self._exec('seekcur', '-15')

    @action
    def status(self):
        """
        :returns: The current state.

        Example response::

            output = {
                "volume": "9",
                "repeat": "0",
                "random": "0",
                "single": "0",
                "consume": "0",
                "playlist": "52",
                "playlistlength": "14",
                "xfade": "0",
                "state": "play",
                "song": "9",
                "songid": "3061",
                "nextsong": "10",
                "nextsongid": "3062",
                "time": "161:255",
                "elapsed": "161.967",
                "bitrate": "320"
            }
        """

        n_tries = 2
        error = None

        while n_tries > 0:
            try:
                n_tries -= 1
                self._connect()
                return self.client.status()
            except Exception as e:
                error = e
                self.logger.warning('Exception while getting MPD status: {}'.
                                    format(str(e)))
                self.client = None

        return None, error

    # noinspection PyTypeChecker
    @action
    def currentsong(self):
        """
        :returns: The currently played track.

        Example response::

            output = {
                "file": "spotify:track:7CO5ADlDN3DcR2pwlnB14P",
                "time": "255",
                "artist": "Elbow",
                "album": "Little Fictions",
                "title": "Kindling",
                "date": "2017",
                "track": "10",
                "pos": "9",
                "id": "3061",
                "albumartist": "Elbow",
                "x-albumuri": "spotify:album:6q5KhDhf9BZkoob7uAnq19"
            }
        """

        track = self._exec('currentsong', return_status=False)
        if 'title' in track and ('artist' not in track
                                 or not track['artist']
                                 or re.search('^https?://', track['file'])
                                 or re.search('^tunein:', track['file'])):
            m = re.match('^\s*(.+?)\s+-\s+(.*)\s*$', track['title'])
            if m and m.group(1) and m.group(2):
                track['artist'] = m.group(1)
                track['title'] = m.group(2)

        return track

    @action
    def playlistinfo(self):
        """
        :returns: The tracks in the current playlist as a list of dicts.

        Example output::

            output = [
                {
                    "file": "spotify:track:79VtgIoznishPUDWO7Kafu",
                    "time": "355",
                    "artist": "Elbow",
                    "album": "Little Fictions",
                    "title": "Trust the Sun",
                    "date": "2017",
                    "track": "3",
                    "pos": "10",
                    "id": "3062",
                    "albumartist": "Elbow",
                    "x-albumuri": "spotify:album:6q5KhDhf9BZkoob7uAnq19"
                },
                {
                    "file": "spotify:track:3EzTre0pxmoMYRuhJKMHj6",
                    "time": "219",
                    "artist": "Elbow",
                    "album": "Little Fictions",
                    "title": "Gentle Storm",
                    "date": "2017",
                    "track": "2",
                    "pos": "11",
                    "id": "3063",
                    "albumartist": "Elbow",
                    "x-albumuri": "spotify:album:6q5KhDhf9BZkoob7uAnq19"
                },
            ]
        """

        return self._exec('playlistinfo', return_status=False)

    @action
    def listplaylists(self):
        """
        :returns: The playlists available on the server as a list of dicts.

        Example response::

            output = [
                {
                    "playlist": "Rock",
                    "last-modified": "2018-06-25T21:28:19Z"
                },
                {
                    "playlist": "Jazz",
                    "last-modified": "2018-06-24T22:28:29Z"
                },
                {
                    # ...
                }
            ]
        """

        return sorted(self._exec('listplaylists', return_status=False),
                      key=lambda p: p['playlist'])

    @action
    def listplaylist(self, name):
        """
        List the items in the specified playlist (without metadata)

        :param name: Name of the playlist
        :type name: str
        """
        return self._exec('listplaylist', name, return_status=False)

    @action
    def listplaylistinfo(self, name):
        """
        List the items in the specified playlist (with metadata)

        :param name: Name of the playlist
        :type name: str
        """
        return self._exec('listplaylistinfo', name, return_status=False)

    @action
    def playlistadd(self, name, uri):
        """
        Add one or multiple resources to a playlist.

        :param name: Playlist name
        :type name: str

        :param uri: URI or path of the resource(s) to be added
        :type uri: str or list[str]
        """

        if isinstance(uri, str):
            uri = [uri]

        for res in uri:
            self._exec('playlistadd', name, res)

    @action
    def playlistdelete(self, name, pos):
        """
        Remove one or multiple tracks from a playlist.

        :param name: Playlist name
        :type name: str

        :param pos: Position or list of positions to remove
        :type pos: int or list[int]
        """

        if isinstance(pos, str):
            pos = int(pos)
        if isinstance(pos, int):
            pos = [pos]

        for p in pos:
            self._exec('playlistdelete', name, p)

    @action
    def playlistmove(self, name, from_pos, to_pos):
        """
        Change the position of a track in the specified playlist

        :param name: Playlist name
        :type name: str

        :param from_pos: Original track position
        :type from_pos: int

        :param to_pos: New track position
        :type to_pos: int
        """
        self._exec('playlistmove', name, from_pos, to_pos)

    @action
    def playlistclear(self, name):
        """
        Clears all the elements from the specified playlist

        :param name: Playlist name
        :type name: str
        """
        self._exec('playlistclear', name)

    @action
    def rename(self, name, new_name):
        """
        Rename a playlist

        :param name: Original playlist name
        :type name: str

        :param new_name: New playlist name
        :type name: str
        """
        self._exec('rename', name, new_name)

    @action
    def lsinfo(self, uri=None):
        """
        Returns the list of playlists and directories on the server
        """

        return self._exec('lsinfo', uri, return_status=False) \
            if uri else self._exec('lsinfo', return_status=False)

    @action
    def plchanges(self, version):
        """
        Show what has changed on the current playlist since a specified playlist
        version number.

        :param version: Version number
        :type version: int

        :returns: A list of dicts representing the songs being added since the specified version
        """

        return self._exec('plchanges', version, return_status=False)

    @action
    def searchaddplaylist(self, name):
        """
        Search and add a playlist by (partial or full) name

        :param name: Playlist name, can be partial
        :type name: str
        """

        playlists = list(map(lambda _: _['playlist'],
                             filter(lambda playlist:
                                    name.lower() in playlist['playlist'].lower(),
                                    self._exec('listplaylists', return_status=False))))

        if len(playlists):
            self._exec('clear')
            self._exec('load', playlists[0])
            self._exec('play')
            return {'playlist': playlists[0]}

    @staticmethod
    def _make_filter(f: dict) -> list:
        ll = []
        for k, v in f.items():
            ll.extend([k, v])
        return ll

    # noinspection PyShadowingBuiltins
    @action
    def find(self, filter: dict, *args, **kwargs):
        """
        Find in the database/library by filter.

        :param filter: Search filter (e.g. ``{"artist": "Led Zeppelin", "album": "IV"}``)
        :returns: list[dict]
        """

        filter = self._make_filter(filter)
        return self._exec('find', *filter, *args, return_status=False, **kwargs)

    # noinspection PyShadowingBuiltins
    @action
    def findadd(self, filter: dict, *args, **kwargs):
        """
        Find in the database/library by filter and add to the current playlist.

        :param filter: Search filter (e.g. ``{"artist": "Led Zeppelin", "album": "IV"}``)
        :returns: list[dict]
        """

        filter = self._make_filter(filter)
        return self._exec('findadd', *filter, *args, return_status=False, **kwargs)

    # noinspection PyShadowingBuiltins
    @action
    def search(self, filter: dict, *args, **kwargs):
        """
        Free search by filter.

        :param filter: Search filter (e.g. ``{"artist": "Led Zeppelin", "album": "IV"}``)
        :returns: list[dict]
        """

        filter = self._make_filter(filter)
        items = self._exec('search', *filter, *args, return_status=False, **kwargs)

        # Spotify results first
        items = sorted(items, key=lambda item:
        0 if item['file'].startswith('spotify:') else 1)

        return items

    # noinspection PyShadowingBuiltins
    @action
    def searchadd(self, filter, *args, **kwargs):
        """
        Free search by filter and add the results to the current playlist.

        :param filter: Search filter (e.g. ``{"artist": "Led Zeppelin", "album": "IV"}``)
        :returns: list[dict]
        """

        filter = self._make_filter(filter)
        return self._exec('searchadd', *filter, *args, return_status=False, **kwargs)


# vim:sw=4:ts=4:et:



import json

from flask import Blueprint, abort, request, Response

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import authenticate, logger, send_message

execute = Blueprint('execute', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    execute,
]


@execute.route('/execute', methods=['POST'])
@authenticate(skip_auth_methods=['session'])
def execute():
    """ Endpoint to execute commands """
    try:
        msg = json.loads(request.data.decode('utf-8'))
    except Exception as e:
        logger().error('Unable to parse JSON from request {}: {}'.format(request.data, str(e)))
        return abort(400, str(e))

    logger().info('Received message on the HTTP backend: {}'.format(msg))

    try:
        response = send_message(msg)
        return Response(str(response or {}), mimetype='application/json')
    except Exception as e:
        logger().error('Error while running HTTP action: {}. Request: {}'.format(str(e), msg))
        return abort(500, str(e))


# vim:sw=4:ts=4:et:



import enum

from platypush.plugins import action
from platypush.plugins.gpio.sensor import GpioSensorPlugin


class MCP3008Mode(enum.Enum):
    SOFTWARE = 'software'
    HARDWARE = 'hardware'


class GpioSensorMcp3008Plugin(GpioSensorPlugin):
    """
    Plugin to read analog sensor values from an MCP3008 chipset.  The MCP3008
    chipset is a circuit that allows you to read measuremnts from multiple
    analog sources (e.g. sensors) and multiplex them to a digital device like a
    Raspberry Pi or a regular laptop.  See
    https://learn.adafruit.com/raspberry-pi-analog-to-digital-converters/mcp3008
    for more info.

    Requires:

        * ``adafruit-mcp3008`` (``pip install adafruit-mcp3008``)
    """

    N_CHANNELS = 8

    def __init__(self, CLK=None, MISO=None, MOSI=None, CS=None, spi_port=None,
                 spi_device=None, channels=None, Vdd=3.3, *args, **kwargs):
        """
        The MCP3008 can be connected in two modes:

            * Hardware SPI mode: advised if you have enough GPIO pins available
                (and slightly faster)
            * Software SPI mode: useful if you don't have all the required GPIO
                PINs for hardware SPI available. Slightly slower, as the conversion
                is done via software, but still relatively performant.

            See
            https://learn.adafruit.com/raspberry-pi-analog-to-digital-converters/mcp3008#wiring
            for info

        :param CLK: (software SPI mode) CLK GPIO PIN
        :type CLK: int
        :param MISO: (software SPI mode) MISO GPIO PIN
        :type MISO: int
        :param MOSI: (software SPI mode) MOSI GPIO PIN
        :type MOSI: int
        :param CS: (software SPI mode) CS GPIO PIN
        :type CS: int

        :param spi_port: (hardware SPI mode) SPI port
        :type spi_port: int
        :param spi_device: (hardware SPI mode) SPI device name
        :type spi_device: str

        :param channels: name-value mapping between MCP3008 output PINs and sensor names. This mapping will be used when you get values through :func:`.get_measurement()`.
            Example::

                channels = {
                    "0": {
                        "name": "temperature",
                        "conv_function": 'round(x*100.0, 2)'  # T = Vout / (10 [mV/C])
                    },
                    "1": {
                        "name": "light",  # ALS-PT9
                        "conv_function": 'round(x*1000.0, 6)'  # ALS-PT9 has a 10 kOhm resistor
                    }
                }

            Note that you can also pass a conversion function as
            ``conv_function`` that will convert the output voltage to whichever
            human-readable value you wish.  In the case above I connected a
            simple temperature sensor to the channel 0 and a simple ALS-PT9
            light sensor to the channel 1, and passed the appropriate conversion
            functions to convert from voltage to, respectively, temperature in
            Celsius degrees and light intensity in lumen. Note that we reference
            the current voltage as ``x`` in ``conv_function``.

        :type channels: dict

        :param Vdd: Input voltage provided to the circuit (default: 3.3V, Raspberry Pi default power source)
        :type Vdd: float
        """

        super().__init__(*args, **kwargs)

        if CLK and MISO and MOSI and CS:
            self.CLK = CLK
            self.MISO = MISO
            self.MOSI = MOSI
            self.CS = CS
            self.mode = MCP3008Mode.SOFTWARE
        elif spi_port and spi_device:
            self.spi_port = spi_port
            self.spi_device = spi_device
            self.mode = MCP3008Mode.HARDWARE
        else:
            raise RuntimeError("At least one mode must be specified.\n" +
                               "Software SPI: Specify CLK, MISO, MOSI and CS pins\n" +
                               "Hardware SPI: Specify spi_port and spi_device\n")

        self.Vdd = Vdd
        self.channels = channels if channels else {}
        self.mcp = None

    def _get_mcp(self):
        import Adafruit_GPIO.SPI as SPI
        import Adafruit_MCP3008

        if self.mode == MCP3008Mode.SOFTWARE:
            self.mcp = Adafruit_MCP3008.MCP3008(clk=self.CLK, cs=self.CS,
                                                miso=self.MISO, mosi=self.MOSI)
        elif self.mode == MCP3008Mode.HARDWARE:
            self.mcp = Adafruit_MCP3008.MCP3008(spi=SPI.SpiDev(self.spi_port, self.spi_device))
        else:
            raise RuntimeError('Unsupported MCP3008 mode: {}'.format(self.mode))

        return self.mcp

    def _convert_to_voltage(self, value):
        return (value * self.Vdd) / 1023.0 if value is not None else None

    @action
    def get_measurement(self):
        """
        Returns a measurement from the sensors connected to the MCP3008 device.
        If channels were passed to the configuration, the appropriate sensor names
        will be used and the voltage will be converted through the appropriate
        conversion function. Example::

            output = {
                "temperature": 21.0,  # Celsius
                "humidity": 45.1    # %
            }

        Otherwise, the output dictionary will contain the channel numbers as key
        and the row voltage (between 0 and 255) will be returned. Example::

            output = {
                "0": 145,
                "1": 130
            }

        """

        mcp = self._get_mcp()
        values = {}

        for i in range(self.N_CHANNELS):
            value = self._convert_to_voltage(mcp.read_adc(i))

            if self.channels:
                if i in self.channels:
                    channel = self.channels[i]
                    if 'conv_function' in channel:
                        x = value
                        value = eval(channel['conv_function'])

                    values[channel['name']] = value
            else:
                values[i] = value

        return values


# vim:sw=4:ts=4:et:



import datetime
import os

from threading import RLock
from typing import Optional, Union

# noinspection PyPackageRequirements
from telegram.ext import Updater
# noinspection PyPackageRequirements
from telegram.message import Message as TelegramMessage
# noinspection PyPackageRequirements
from telegram.user import User as TelegramUser

from platypush.message.response.chat.telegram import TelegramMessageResponse, TelegramFileResponse, \
    TelegramChatResponse, TelegramUserResponse, TelegramUsersResponse
from platypush.plugins import Plugin, action


class Resource:
    def __init__(self, file_id: Optional[int] = None, url: Optional[str] = None, path: Optional[str] = None):
        assert file_id or url or path, 'You need to specify either file_id, url or path'
        self.file_id = file_id
        self.url = url
        self.path = path
        self._file = None

    def __enter__(self):
        if self.path:
            self._file = open(os.path.abspath(os.path.expanduser(self.path)), 'rb')
            return self._file

        return self.file_id or self.url

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._file:
            self._file.close()


class ChatTelegramPlugin(Plugin):
    """
    Plugin to programmatically send Telegram messages through a Telegram bot. In order to send messages to contacts,
    groups or channels you'll first need to register a bot. To do so:

        1. Open a Telegram conversation with the `@BotFather <https://telegram.me/BotFather>`_.
        2. Send ``/start`` followed by ``/newbot``. Choose a display name and a username for your bot.
        3. Copy the provided API token in the configuration of this plugin.
        4. Open a conversation with your newly created bot.

    Requires:

        * **python-telegram-bot** (``pip install python-telegram-bot``)

    """

    def __init__(self, api_token: str, **kwargs):
        """
        :param api_token: API token as returned by the `@BotFather <https://telegram.me/BotFather>`_
        """
        super().__init__(**kwargs)

        self._api_token = api_token
        self._telegram_lock = RLock()
        self._telegram: Optional[Updater] = None

    def get_telegram(self) -> Updater:
        with self._telegram_lock:
            if self._telegram:
                return self._telegram

            self._telegram = Updater(self._api_token, use_context=True)
            return self._telegram

    @staticmethod
    def parse_msg(msg: TelegramMessage) -> TelegramMessageResponse:
        return TelegramMessageResponse(
            message_id=msg.message_id,
            chat_id=msg.chat_id,
            chat_username=msg.chat.username,
            chat_firstname=msg.chat.first_name,
            chat_lastname=msg.chat.last_name,
            from_user_id=msg.from_user.id if msg.from_user else None,
            from_username=msg.from_user.username if msg.from_user else None,
            from_firstname=msg.from_user.first_name if msg.from_user else None,
            from_lastname=msg.from_user.last_name if msg.from_user else None,
            text=msg.text,
            caption=msg.caption,
            creation_date=msg.date,
            edit_date=msg.edit_date,
            forward_date=msg.forward_date,
            forward_from_message_id=msg.forward_from_message_id,
            photo_file_id=msg.photo[0].file_id if msg.photo else None,
            photo_file_size=msg.photo[0].file_size if msg.photo else None,
            photo_width=msg.photo[0].width if msg.photo else None,
            photo_height=msg.photo[0].height if msg.photo else None,
            document_file_id=msg.document.file_id if msg.document else None,
            document_file_name=msg.document.file_name if msg.document else None,
            document_file_size=msg.document.file_size if msg.document else None,
            document_mime_type=msg.document.mime_type if msg.document else None,
            audio_file_id=msg.audio.file_id if msg.audio else None,
            audio_file_size=msg.audio.file_size if msg.audio else None,
            audio_performer=msg.audio.performer if msg.audio else None,
            audio_title=msg.audio.title if msg.audio else None,
            audio_duration=msg.audio.duration if msg.audio else None,
            audio_mime_type=msg.audio.mime_type if msg.audio else None,
            video_file_id=msg.video.file_id if msg.video else None,
            video_file_size=msg.video.file_size if msg.video else None,
            video_duration=msg.video.duration if msg.video else None,
            video_width=msg.video.width if msg.video else None,
            video_height=msg.video.height if msg.video else None,
            video_mime_type=msg.video.mime_type if msg.video else None,
            location_latitude=msg.location.latitude if msg.location else None,
            location_longitude=msg.location.longitude if msg.location else None,
            contact_phone_number=msg.contact.phone_number if msg.contact else None,
            contact_first_name=msg.contact.first_name if msg.contact else None,
            contact_last_name=msg.contact.last_name if msg.contact else None,
            contact_user_id=msg.contact.user_id if msg.contact else None,
            contact_vcard=msg.contact.vcard if msg.contact else None,
            link=msg.link,
            media_group_id=msg.media_group_id
        )

    @staticmethod
    def parse_user(user: TelegramUser) -> TelegramUserResponse:
        return TelegramUserResponse(
            user_id=user.id,
            username=user.username,
            is_bot=user.is_bot,
            first_name=user.first_name,
            last_name=user.last_name,
            language_code=user.language_code,
            link=user.link
        )

    @action
    def send_message(self, chat_id: Union[str, int], text: str, parse_mode: Optional[str] = None,
                     disable_web_page_preview: bool = False, disable_notification: bool = False,
                     reply_to_message_id: Optional[int] = None) -> TelegramMessageResponse:
        """
        Send a message to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param text: Text to be sent.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown or HTML content.
        :param disable_web_page_preview: If True then web previews for URLs will be disabled.
        :param disable_notification: If True then no notification will be sent to the users.
        :param reply_to_message_id: If set then the message will be sent as a response to the specified message.
        """

        telegram = self.get_telegram()
        msg = telegram.bot.send_message(chat_id=chat_id,
                                        text=text,
                                        parse_mode=parse_mode,
                                        disable_web_page_preview=disable_web_page_preview,
                                        disable_notification=disable_notification,
                                        reply_to_message_id=reply_to_message_id)

        return self.parse_msg(msg)

    @action
    def send_photo(self, chat_id: Union[str, int],
                   file_id: Optional[int] = None,
                   url: Optional[str] = None,
                   path: Optional[str] = None,
                   caption: Optional[str] = None,
                   parse_mode: Optional[str] = None,
                   disable_notification: bool = False,
                   reply_to_message_id: Optional[int] = None,
                   timeout: int = 20) -> TelegramMessageResponse:
        """
        Send a picture to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param file_id: Set it if the file already exists on Telegram servers and has a file_id.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param url: Set it if you want to send a file from a remote URL.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param path: Set it if you want to send a file from the local filesystem.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param caption: Optional caption for the picture.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown or HTML content.
        :param disable_notification: If True then no notification will be sent to the users.
        :param reply_to_message_id: If set then the message will be sent as a response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()

        with Resource(file_id=file_id, url=url, path=path) as resource:
            msg = telegram.bot.send_photo(chat_id=chat_id,
                                          photo=resource,
                                          caption=caption,
                                          disable_notification=disable_notification,
                                          reply_to_message_id=reply_to_message_id,
                                          timeout=timeout, parse_mode=parse_mode)

        return self.parse_msg(msg)

    @action
    def send_audio(self, chat_id: Union[str, int],
                   file_id: Optional[int] = None,
                   url: Optional[str] = None,
                   path: Optional[str] = None,
                   caption: Optional[str] = None,
                   performer: Optional[str] = None,
                   title: Optional[str] = None,
                   duration: Optional[float] = None,
                   parse_mode: Optional[str] = None,
                   disable_notification: bool = False,
                   reply_to_message_id: Optional[int] = None,
                   timeout: int = 20) -> TelegramMessageResponse:
        """
        Send audio to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param file_id: Set it if the file already exists on Telegram servers and has a file_id.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param url: Set it if you want to send a file from a remote URL.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param path: Set it if you want to send a file from the local filesystem.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param caption: Optional caption for the picture.
        :param performer: Optional audio performer.
        :param title: Optional audio title.
        :param duration: Duration of the audio in seconds.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown or HTML content.
        :param disable_notification: If True then no notification will be sent to the users.
        :param reply_to_message_id: If set then the message will be sent as a response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()

        with Resource(file_id=file_id, url=url, path=path) as resource:
            msg = telegram.bot.send_audio(chat_id=chat_id,
                                          audio=resource,
                                          caption=caption,
                                          disable_notification=disable_notification,
                                          performer=performer,
                                          title=title,
                                          duration=duration,
                                          reply_to_message_id=reply_to_message_id,
                                          timeout=timeout,
                                          parse_mode=parse_mode)

        return self.parse_msg(msg)

    @action
    def send_document(self, chat_id: Union[str, int],
                      file_id: Optional[int] = None,
                      url: Optional[str] = None,
                      path: Optional[str] = None,
                      filename: Optional[str] = None,
                      caption: Optional[str] = None,
                      parse_mode: Optional[str] = None,
                      disable_notification: bool = False,
                      reply_to_message_id: Optional[int] = None,
                      timeout: int = 20) -> TelegramMessageResponse:
        """
        Send a document to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param file_id: Set it if the file already exists on Telegram servers and has a file_id.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param url: Set it if you want to send a file from a remote URL.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param path: Set it if you want to send a file from the local filesystem.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param filename: Name of the file as it will be shown in Telegram.
        :param caption: Optional caption for the picture.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown or HTML content.
        :param disable_notification: If True then no notification will be sent to the users.
        :param reply_to_message_id: If set then the message will be sent as a response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()

        with Resource(file_id=file_id, url=url, path=path) as resource:
            msg = telegram.bot.send_document(chat_id=chat_id,
                                             document=resource,
                                             filename=filename,
                                             caption=caption,
                                             disable_notification=disable_notification,
                                             reply_to_message_id=reply_to_message_id,
                                             timeout=timeout,
                                             parse_mode=parse_mode)

        return self.parse_msg(msg)

    @action
    def send_video(self, chat_id: Union[str, int],
                   file_id: Optional[int] = None,
                   url: Optional[str] = None,
                   path: Optional[str] = None,
                   duration: Optional[int] = None,
                   caption: Optional[str] = None,
                   width: Optional[int] = None,
                   height: Optional[int] = None,
                   parse_mode: Optional[str] = None,
                   disable_notification: bool = False,
                   reply_to_message_id: Optional[int] = None,
                   timeout: int = 20) -> TelegramMessageResponse:
        """
        Send a video to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param file_id: Set it if the file already exists on Telegram servers and has a file_id.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param url: Set it if you want to send a file from a remote URL.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param path: Set it if you want to send a file from the local filesystem.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param duration: Duration in seconds.
        :param caption: Optional caption for the picture.
        :param width: Video width.
        :param height: Video height.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown or HTML content.
        :param disable_notification: If True then no notification will be sent to the users.
        :param reply_to_message_id: If set then the message will be sent as a response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()

        with Resource(file_id=file_id, url=url, path=path) as resource:
            msg = telegram.bot.send_video(chat_id=chat_id,
                                          video=resource,
                                          duration=duration,
                                          caption=caption,
                                          width=width,
                                          height=height,
                                          disable_notification=disable_notification,
                                          reply_to_message_id=reply_to_message_id,
                                          timeout=timeout,
                                          parse_mode=parse_mode)

        return self.parse_msg(msg)

    @action
    def send_animation(self, chat_id: Union[str, int],
                       file_id: Optional[int] = None,
                       url: Optional[str] = None,
                       path: Optional[str] = None,
                       duration: Optional[int] = None,
                       caption: Optional[str] = None,
                       width: Optional[int] = None,
                       height: Optional[int] = None,
                       parse_mode: Optional[str] = None,
                       disable_notification: bool = False,
                       reply_to_message_id: Optional[int] = None,
                       timeout: int = 20) -> TelegramMessageResponse:
        """
        Send an animation (GIF or H.264/MPEG-4 AVC video without sound) to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param file_id: Set it if the file already exists on Telegram servers and has a file_id.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param url: Set it if you want to send a file from a remote URL.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param path: Set it if you want to send a file from the local filesystem.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param duration: Duration in seconds.
        :param caption: Optional caption for the picture.
        :param width: Video width.
        :param height: Video height.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown or HTML content.
        :param disable_notification: If True then no notification will be sent to the users.
        :param reply_to_message_id: If set then the message will be sent as a response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()

        with Resource(file_id=file_id, url=url, path=path) as resource:
            msg = telegram.bot.send_animation(chat_id=chat_id,
                                              animation=resource,
                                              duration=duration,
                                              caption=caption,
                                              width=width,
                                              height=height,
                                              disable_notification=disable_notification,
                                              reply_to_message_id=reply_to_message_id,
                                              timeout=timeout,
                                              parse_mode=parse_mode)

        return self.parse_msg(msg)

    @action
    def send_voice(self, chat_id: Union[str, int],
                   file_id: Optional[int] = None,
                   url: Optional[str] = None,
                   path: Optional[str] = None,
                   caption: Optional[str] = None,
                   duration: Optional[float] = None,
                   parse_mode: Optional[str] = None,
                   disable_notification: bool = False,
                   reply_to_message_id: Optional[int] = None,
                   timeout: int = 20) -> TelegramMessageResponse:
        """
        Send audio to a chat as a voice file. For this to work, your audio must be in an .ogg file encoded with OPUS
        (other formats may be sent as Audio or Document).

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param file_id: Set it if the file already exists on Telegram servers and has a file_id.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param url: Set it if you want to send a file from a remote URL.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param path: Set it if you want to send a file from the local filesystem.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param caption: Optional caption for the picture.
        :param duration: Duration of the voice in seconds.
        :param parse_mode: Set to 'Markdown' or 'HTML' to send either Markdown or HTML content.
        :param disable_notification: If True then no notification will be sent to the users.
        :param reply_to_message_id: If set then the message will be sent as a response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()

        with Resource(file_id=file_id, url=url, path=path) as resource:
            msg = telegram.bot.send_voice(chat_id=chat_id,
                                          voice=resource,
                                          caption=caption,
                                          disable_notification=disable_notification,
                                          duration=duration,
                                          reply_to_message_id=reply_to_message_id,
                                          timeout=timeout, parse_mode=parse_mode)

        return self.parse_msg(msg)

    @action
    def send_video_note(self, chat_id: Union[str, int],
                        file_id: Optional[int] = None,
                        url: Optional[str] = None,
                        path: Optional[str] = None,
                        duration: Optional[int] = None,
                        disable_notification: bool = False,
                        reply_to_message_id: Optional[int] = None,
                        timeout: int = 20) -> TelegramMessageResponse:
        """
        Send a video note to a chat. As of v.4.0, Telegram clients support rounded square mp4 videos of up to
        1 minute long.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param file_id: Set it if the file already exists on Telegram servers and has a file_id.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param url: Set it if you want to send a file from a remote URL.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param path: Set it if you want to send a file from the local filesystem.
            Note that you'll have to specify either ``file_id``, ``url`` or ``path``.

        :param duration: Duration in seconds.
        :param disable_notification: If True then no notification will be sent to the users.
        :param reply_to_message_id: If set then the message will be sent as a response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()

        with Resource(file_id=file_id, url=url, path=path) as resource:
            msg = telegram.bot.send_video_note(chat_id=chat_id,
                                               video=resource,
                                               duration=duration,
                                               disable_notification=disable_notification,
                                               reply_to_message_id=reply_to_message_id,
                                               timeout=timeout)

        return self.parse_msg(msg)

    @action
    def send_location(self, chat_id: Union[str, int],
                      latitude: float,
                      longitude: float,
                      disable_notification: bool = False,
                      reply_to_message_id: Optional[int] = None,
                      timeout: int = 20) -> TelegramMessageResponse:
        """
        Send a location to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param latitude: Latitude
        :param longitude: Longitude
        :param disable_notification: If True then no notification will be sent to the users.
        :param reply_to_message_id: If set then the message will be sent as a response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()
        msg = telegram.bot.send_location(chat_id=chat_id,
                                         latitude=latitude,
                                         longitude=longitude,
                                         disable_notification=disable_notification,
                                         reply_to_message_id=reply_to_message_id,
                                         timeout=timeout)

        return self.parse_msg(msg)

    @action
    def send_venue(self, chat_id: Union[str, int],
                   latitude: float,
                   longitude: float,
                   title: str,
                   address: str,
                   foursquare_id: Optional[str] = None,
                   foursquare_type: Optional[str] = None,
                   disable_notification: bool = False,
                   reply_to_message_id: Optional[int] = None,
                   timeout: int = 20) -> TelegramMessageResponse:
        """
        Send the address of a venue to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param latitude: Latitude
        :param longitude: Longitude
        :param title: Venue name.
        :param address: Venue address.
        :param foursquare_id: Foursquare ID.
        :param foursquare_type: Foursquare type.
        :param disable_notification: If True then no notification will be sent to the users.
        :param reply_to_message_id: If set then the message will be sent as a response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()
        msg = telegram.bot.send_venue(chat_id=chat_id,
                                      latitude=latitude,
                                      longitude=longitude,
                                      title=title,
                                      address=address,
                                      foursquare_id=foursquare_id,
                                      foursquare_type=foursquare_type,
                                      disable_notification=disable_notification,
                                      reply_to_message_id=reply_to_message_id,
                                      timeout=timeout)

        return self.parse_msg(msg)

    @action
    def send_contact(self, chat_id: Union[str, int],
                     phone_number: str,
                     first_name: str,
                     last_name: Optional[str] = None,
                     vcard: Optional[str] = None,
                     disable_notification: bool = False,
                     reply_to_message_id: Optional[int] = None,
                     timeout: int = 20) -> TelegramMessageResponse:
        """
        Send a contact to a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param phone_number: Phone number.
        :param first_name: First name.
        :param last_name: Last name.
        :param vcard: Additional contact info in vCard format (0-2048 bytes).
        :param disable_notification: If True then no notification will be sent to the users.
        :param reply_to_message_id: If set then the message will be sent as a response to the specified message.
        :param timeout: Upload timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()
        msg = telegram.bot.send_contact(chat_id=chat_id,
                                        phone_number=phone_number,
                                        first_name=first_name,
                                        last_name=last_name,
                                        vcard=vcard,
                                        disable_notification=disable_notification,
                                        reply_to_message_id=reply_to_message_id,
                                        timeout=timeout)

        return self.parse_msg(msg)

    @action
    def get_file(self, file_id: str, timeout: int = 20) -> TelegramFileResponse:
        """
        Get the info and URL of an uploaded file by file_id.

        :param file_id: File ID.
        :param timeout: Upload timeout (default: 20 seconds).
        """

        telegram = self.get_telegram()
        file = telegram.bot.get_file(file_id, timeout=timeout)
        return TelegramFileResponse(file_id=file.file_id, file_path=file.file_path, file_size=file.file_size)

    @action
    def get_chat(self, chat_id: Union[int, str], timeout: int = 20) -> TelegramChatResponse:
        """
        Get the info about a Telegram chat.

        :param chat_id: Chat ID.
        :param timeout: Upload timeout (default: 20 seconds).
        """

        telegram = self.get_telegram()
        chat = telegram.bot.get_chat(chat_id, timeout=timeout)
        return TelegramChatResponse(chat_id=chat.id,
                                    link=chat.link,
                                    username=chat.username,
                                    invite_link=chat.invite_link,
                                    title=chat.title,
                                    description=chat.description,
                                    type=chat.type,
                                    first_name=chat.first_name,
                                    last_name=chat.last_name)

    @action
    def get_chat_user(self, chat_id: Union[int, str], user_id: int, timeout: int = 20) -> TelegramUserResponse:
        """
        Get the info about a user connected to a chat.

        :param chat_id: Chat ID.
        :param user_id: User ID.
        :param timeout: Upload timeout (default: 20 seconds).
        """

        telegram = self.get_telegram()
        user = telegram.bot.get_chat_member(chat_id, user_id, timeout=timeout)
        return TelegramUserResponse(user_id=user.user.id,
                                    link=user.user.link,
                                    username=user.user.username,
                                    first_name=user.user.first_name,
                                    last_name=user.user.last_name,
                                    is_bot=user.user.is_bot,
                                    language_code=user.user.language_code)

    @action
    def get_chat_administrators(self, chat_id: Union[int, str], timeout: int = 20) -> TelegramUsersResponse:
        """
        Get the list of the administrators of a chat.

        :param chat_id: Chat ID.
        :param timeout: Upload timeout (default: 20 seconds).
        """

        telegram = self.get_telegram()
        admins = telegram.bot.get_chat_administrators(chat_id, timeout=timeout)
        return TelegramUsersResponse([
            TelegramUserResponse(
                user_id=user.user.id,
                link=user.user.link,
                username=user.user.username,
                first_name=user.user.first_name,
                last_name=user.user.last_name,
                is_bot=user.user.is_bot,
                language_code=user.user.language_code,
            ) for user in admins
        ])

    @action
    def get_chat_members_count(self, chat_id: Union[int, str], timeout: int = 20) -> int:
        """
        Get the number of users in a chat.

        :param chat_id: Chat ID.
        :param timeout: Upload timeout (default: 20 seconds).
        """
        telegram = self.get_telegram()
        return telegram.bot.get_chat_members_count(chat_id, timeout=timeout)

    @action
    def kick_chat_member(self, chat_id: Union[str, int],
                         user_id: int,
                         until_date: Optional[datetime.datetime] = None,
                         timeout: int = 20):
        """
        Kick a user from a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param user_id: Unique user ID.
        :param until_date: End date for the ban.
        :param timeout: Request timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()
        telegram.bot.kick_chat_member(
            chat_id=chat_id,
            user_id=user_id,
            until_date=until_date,
            timeout=timeout)

    @action
    def unban_chat_member(self, chat_id: Union[str, int],
                          user_id: int,
                          timeout: int = 20):
        """
        Lift the ban from a chat member.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param user_id: Unique user ID.
        :param timeout: Request timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()
        telegram.bot.unban_chat_member(
            chat_id=chat_id,
            user_id=user_id,
            timeout=timeout)

    @action
    def promote_chat_member(self, chat_id: Union[str, int],
                            user_id: int,
                            can_change_info: Optional[bool] = None,
                            can_post_messages: Optional[bool] = None,
                            can_edit_messages: Optional[bool] = None,
                            can_delete_messages: Optional[bool] = None,
                            can_invite_users: Optional[bool] = None,
                            can_restrict_members: Optional[bool] = None,
                            can_promote_members: Optional[bool] = None,
                            can_pin_messages: Optional[bool] = None,
                            timeout: int = 20):
        """
        Promote or demote a member.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param user_id: Unique user ID.
        :param can_change_info: Pass True if the user can change channel info.
        :param can_post_messages: Pass True if the user can post messages.
        :param can_edit_messages: Pass True if the user can edit messages.
        :param can_delete_messages: Pass True if the user can delete messages.
        :param can_invite_users: Pass True if the user can invite other users to the channel/group.
        :param can_restrict_members: Pass True if the user can restrict the permissions of other users.
        :param can_promote_members: Pass True if the user can promote mebmers.
        :param can_pin_messages: Pass True if the user can pin messages.
        :param timeout: Request timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()
        telegram.bot.promote_chat_member(
            chat_id=chat_id,
            user_id=user_id,
            can_change_info=can_change_info,
            can_post_messages=can_post_messages,
            can_edit_messages=can_edit_messages,
            can_delete_messages=can_delete_messages,
            can_invite_users=can_invite_users,
            can_restrict_members=can_restrict_members,
            can_promote_members=can_promote_members,
            can_pin_messages=can_pin_messages,
            timeout=timeout)

    @action
    def set_chat_title(self, chat_id: Union[str, int],
                       title: str,
                       timeout: int = 20):
        """
        Set the title of a channel/group.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param title: New chat title.
        :param timeout: Request timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()
        telegram.bot.set_chat_title(
            chat_id=chat_id,
            description=title,
            timeout=timeout)

    @action
    def set_chat_description(self, chat_id: Union[str, int],
                             description: str,
                             timeout: int = 20):
        """
        Set the description of a channel/group.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param description: New chat description.
        :param timeout: Request timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()
        telegram.bot.set_chat_description(
            chat_id=chat_id,
            description=description,
            timeout=timeout)

    @action
    def set_chat_photo(self, chat_id: Union[str, int],
                       path: str,
                       timeout: int = 20):
        """
        Set the photo of a channel/group.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param path: Path of the new image.
        :param timeout: Request timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()

        with Resource(path=path) as resource:
            telegram.bot.set_chat_photo(
                chat_id=chat_id,
                photo=resource,
                timeout=timeout)

    @action
    def delete_chat_photo(self, chat_id: Union[str, int],
                          timeout: int = 20):
        """
        Delete the photo of a channel/group.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param timeout: Request timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()
        telegram.bot.delete_chat_photo(
            chat_id=chat_id,
            timeout=timeout)

    @action
    def pin_chat_message(self, chat_id: Union[str, int],
                         message_id: int,
                         disable_notification: Optional[bool] = None,
                         timeout: int = 20):
        """
        Pin a message in a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param message_id: Message ID.
        :param disable_notification: If True then no notification will be sent to the users.
        :param timeout: Request timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()
        telegram.bot.pin_chat_message(
            chat_id=chat_id,
            message_id=message_id,
            disable_notification=disable_notification,
            timeout=timeout)

    @action
    def unpin_chat_message(self, chat_id: Union[str, int],
                           timeout: int = 20):
        """
        Unpin the message of a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param timeout: Request timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()
        telegram.bot.unpin_chat_message(
            chat_id=chat_id,
            timeout=timeout)

    @action
    def leave_chat(self, chat_id: Union[str, int],
                   timeout: int = 20):
        """
        Leave a chat.

        :param chat_id: Chat ID. Can be either a numerical ID or a unique identifier in the format ``@channelname``.
            In order to get your own Telegram chat_id open a conversation with
            `@IDBot <https://telegram.me/IDBot>`_ and type ``/start`` followed by ``/getid``. Similar procedures
            also exist to get a group or channel chat_id - just Google for "Telegram get channel/group chat_id".

        :param timeout: Request timeout (default: 20 seconds)
        """

        telegram = self.get_telegram()
        telegram.bot.leave_chat(
            chat_id=chat_id,
            timeout=timeout)


# vim:sw=4:ts=4:et:
